/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: tap.c,v 4.27 2000/09/24 10:10:51 erik Rel $";
#endif
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include "cpu.h"
#include "debug.h"
#include "resource.h"
#include "machine.h"
#include "util.h"
#include "tape.h"
#include "tzx.h"
#include "tap.h"
#ifdef DEBUG
#define DEB(x)							{ if (CFGGET(debug) & D_TAPE) { x } }
#else
#define DEB(x)
#endif
int
TapInAttach(char *fname, uns8 **begin, uns8 **end)
{
	int readonly;
	size_t size;
	void *mmp;
	uns8 *scanBegin, *scanEnd, *scanPtr;
	unsigned int length;
	char pattern[] = "xzxXXXXXX", *tmpfname = (char *)0;
	FILE *fp = (FILE *)0;
	int ret = -1;
	readonly = 1;
	mmp = Mmap(fname, 1, &readonly, &size);
	if (mmp == (void *)-1)
	{
		goto quit;
	}
	if ((tmpfname = mktemp(pattern)) == (char *)0 || (fp = Fopen(tmpfname, "wb")) == (FILE *)0)
	{
		goto quit;
	}
	if (fwrite("ZXTape!\032\001\012", 1, (size_t)10, fp) != 10)
	{
		goto quit;
	}
	scanBegin = (uns8 *)mmp;
	scanEnd = (uns8 *)mmp + size;
	for (scanPtr = scanBegin; scanPtr < scanEnd; scanPtr += length)
	{
		length = (unsigned int)GET2(scanPtr, 0);
		scanPtr += 2;
		DEB(Msg(M_DEBUG, "TAP: scaning, id = %02x, pos = %6d, block = %5d",
				length != 0 ? *scanPtr : 0xff, scanPtr - 2 - scanBegin, length););
		if (scanPtr >= scanEnd)
		{
			Msg(M_ERR, "premature end on <%s>", fname);
			break;
		}
		if (scanPtr + length > scanEnd)
		{
			Msg(M_WARN, "truncating last TAP block");
			length = (unsigned int)(scanEnd - scanPtr);
		}
		(void)fputc(0x10, fp);
		(void)fputc(0xe8, fp);
		(void)fputc(0x03, fp);
		(void)fputc((int)(length & 0xff), fp);
		(void)fputc((int)(length >> 8), fp);
		if (fwrite(scanPtr, 1, (size_t)length, fp) != (size_t)length)
		{
			goto quit;
		}
	}
	(void)fflush(fp);
	ret = TzxInAttach(tmpfname, begin, end);
  quit:
	if (fp != (FILE *)0)
	{
		(void)fclose(fp);
	}
	if (tmpfname != (char *)0)
	{
		unlink(tmpfname);
	}
	if (mmp != (void *)-1)
	{
		Munmap(mmp, size);
	}
	return ret;
}
#ifndef NO_ED_TRAPS
void
TapInReadBlock(uns8 *data, int len)
{
	uns8 byte, crc;
	uns16 swap;
	int error = 0;
	DEB(Msg(M_DEBUG, "TAP: %s, id = %02x, block = %5d", F1 & FLAG_C ? "loading" : "verifying", (int)A1, len););
	if (len != 0)
	{
		{
			int de = (DE != 0 ? (int)DE : 65636) + 2;
			len = de > len ? len : de;
		}
		crc = *data++;
		len--;
		for (error = 1; ; DE--, IX++)
		{
			byte = *data++;
			crc ^= byte;
			if (--len == 0)
			{
				if (crc == 0)
				{
					error = 0;
				}
#ifdef DEBUG
				else if ((CFGGET(debug) & D_TAPE) != 0)
				{
					DEB(Msg(M_DEBUG, "TAP: checksum error"););
				}
#endif
				break;
			}
			if ((F1 & FLAG_C) != 0)
			{
				Machine->writeMem(IX, byte);
			}
			else
			{
				if (byte != Machine->readMem(IX))
				{
					break;
				}
			}
		}
	}
	if (error != 0)
	{
		F1 &= ~(FLAG_Z | FLAG_C);
	}
	else
	{
		F1 |= FLAG_Z | FLAG_C;
	}
	swap = AF; AF = AF1; AF1 = swap;
	RET();
}
#endif
#ifndef NO_ED_TRAPS
void
TapOutWriteBlock(FILE *fp)
{
	uns16 length;
	int byte, crc;
	length = DE + 2;
	(void)putc(length & 0xff, fp);
	(void)putc(length >> 8, fp);
	(void)putc((int)A, fp);
	crc = A;
	while (DE-- != 0)
	{
		byte = Machine->readMem(IX);
		(void)putc(byte, fp);
		IX++;
		crc ^= byte;
	}
	(void)putc(crc, fp);
	(void)fflush(fp);
	F |= FLAG_C;
	IX += DE;
	DE = 0;
	RET();
}
#endif
