/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifdef XZX_DIDAKTIK
#ifndef lint
static char rcsid[] = "$Id: didaktik.c,v 1.6 2000/11/18 18:45:50 erik Rel $";
#endif
#include <assert.h>
#include "cpu.h"
#include "debug.h"
#include "resource.h"
#include "mem.h"
#include "io.h"
#include "machine.h"
#include "util.h"
#ifdef AUDIO
#include "audio.h"
#endif
#include "emul.h"
#include "keyboard.h"
#include "screen.h"
#include "tables.h"
#include "tape.h"
#include "didaktik.h"
#define Y_INTERRUPT						0
#define Y_RETRACE						16
#define Y_TOP_BORDER					48
#define Y_SCREEN						192
#define Y_BOTTOM_BORDER					56
#define X_LEFT_BORDER					24
#define X_SCREEN						128
#define X_RIGHT							24
#define X_RETRACE						48
#define ULA_DRAWS_SCREEN				0x00
#define ULA_DRAWS_BORDER				0x20
#define ULA_HORIZONTAL_RETRACE			0x40
#define ULA_VERTICAL_RETRACE			0x80
#define ULA_DELAY_MASK					0x0f
enum { ROM_UNKNOWN, ROM_DIDAKTIK93 };
#ifdef DEBUG
#define DEB(x)							{ if ((CFGGET(debug) & D_IO) != 0) { x } }
#else
#define DEB(x)
#endif
static void didaktikReset(void);
static void didaktikNmi(void);
static uns8 didaktikReadOpcode(uns16);
static uns8 didaktikReadMem(uns16);
static void didaktikWriteMem(uns16, uns8);
static void didaktikInitPages(void);
static void didaktikPageIn(int, int);
static void didaktikOutPort(uns16, uns8);
static uns8 didaktikInPort(uns16);
int DidaktikInsertDisk(int i) {	return 0; }
void DidaktikEjectDisk(int i) { }
char *DidaktikDiskName(int i) { return (char *)0; }
void
DidaktikInit(void)
{
	MachineSettings *msp;
	int checksum, romtype;
	unsigned int i, x, y;
	msp = &Machines[SP_DIDAKTIK];
	msp->name = "Didaktik M";
	msp->basicRom = DIDROM0;
	msp->tvLines = Y_RETRACE + Y_TOP_BORDER + Y_SCREEN + Y_BOTTOM_BORDER;
	msp->tstatesPerLine = X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE;
	msp->tstatesPerFrame = msp->tvLines * msp->tstatesPerLine;
	msp->intLine = Y_INTERRUPT;
	msp->topBorder = Y_RETRACE + Y_TOP_BORDER;
	msp->ulaDelay = Malloc(msp->tstatesPerFrame, "DidaktikInit");
	msp->reset = didaktikReset;
	msp->nmi = didaktikNmi;
	msp->readOpcode = didaktikReadOpcode;
	msp->readMem = didaktikReadMem;
	msp->writeMem = didaktikWriteMem;
	msp->inPort = didaktikInPort;
	msp->outPort = didaktikOutPort;
	msp->initPages = didaktikInitPages;
	msp->pageIn = didaktikPageIn;
	if (!LoadRomImage(CFGGET(didaktikRom), DIDROM0, 2))
	{
		checksum = GetRomChecksum(msp->basicRom, XZX_PAGESIZE);
		switch (checksum)
		{
			case 34419:
				romtype = ROM_DIDAKTIK93;
				break;
			default:
				Msg(M_WARN, "unknown Didaktik ROM image - checksum %u",
					checksum);
				romtype = ROM_UNKNOWN;
				break;
		}
#if 0
#ifdef NO_ED_TRAPS
		if (romtype == ROM_DIDAKTIK93)
		{
			RealMemory[msp->basicRom][0x0562] = 0xed;
			RealMemory[msp->basicRom][0x0563] = LOAD_BPT;
			RealMemory[msp->basicRom][0x04c6] = 0xed;
			RealMemory[msp->basicRom][0x04c7] = SAVE_BPT;
#ifdef XBELL_AUDIO
			RealMemory[msp->basicRom][0x03b5] = 0xed;
			RealMemory[msp->basicRom][0x03b6] = BEEP_BPT;
#endif
		}
#endif
#endif
	}
	for (i = 0, y = 0; y < Y_RETRACE; y++)
	{
		for (x = 0; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_VERTICAL_RETRACE;
		}
	}
	for (; y < Y_RETRACE + Y_TOP_BORDER; y++)
	{
		for (x = 0; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_HORIZONTAL_RETRACE;
		}
	}
	for (; y < Y_RETRACE + Y_TOP_BORDER + Y_SCREEN; y++)
	{
		for (x = 0; x < X_LEFT_BORDER; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_SCREEN;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_HORIZONTAL_RETRACE;
		}
	}
	for (; y < Y_RETRACE + Y_TOP_BORDER + Y_SCREEN + Y_BOTTOM_BORDER; y++)
	{
		for (x = 0; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_HORIZONTAL_RETRACE;
		}
	}
	assert(i == msp->tstatesPerFrame);
}
static void
didaktikReset(void)
{
	Z80_Reset();
	Machine->initPages();
	ScreenSelect = RAM5;
	Last0xFFFD = 0;
	(void)memset(Psg, 0x00, sizeof(Psg));
#ifdef REAL_AUDIO
	AyReset();
#endif
}
static void
didaktikNmi(void)
{
	Z80_NMI((NmiCallbackFunc)0);
}
static uns8
didaktikReadOpcode(uns16 addr)
{
	if (addr == 0x000 || addr == 0x0008)
	{
		Msg(M_DEBUG, "didrom1 ein");
		Machine->pageIn(0, DIDROM1);
	}
	else if (addr == 0x1700)
	{
		Msg(M_DEBUG, "didrom0 ein");
		Machine->pageIn(0, DIDROM0);
	}
	return MEM(addr);
}
static uns8
didaktikReadMem(uns16 addr)
{
	return MEM(addr);
}
static void
didaktikWriteMem(uns16 addr, uns8 byte)
{
	if (!READ_ONLY(addr) || (RPAGE(addr) == DIDROM1 && addr >= 0x3800))
	{
		MEM(addr) = byte;
		if (RPAGE(addr) == RAM5 && (addr &= 0x3fff) < (uns16)(PIXEL_LENGTH + ATTR_LENGTH))
		{
			ScreenMarkDirty(addr);
		}
	}
}
static void
didaktikInitPages(void)
{
	Machine->pageIn(0, DIDROM0);
	Machine->pageIn(1, RAM5);
	Machine->pageIn(2, RAM2);
	Machine->pageIn(3, RAM0);
}
static void
didaktikPageIn(int z80page, int vpage)
{
	int i = z80page << 1;
	TheMemory[i] = RealMemory[vpage];
	RealPageNumbers[i] = vpage;
	i++;
	TheMemory[i] = RealMemory[vpage] + XZX_PAGESIZE / 2;
	RealPageNumbers[i] = vpage;
}
static void
didaktikOutPort(uns16 port, uns8 val)
{
	DEB(Msg(M_DEBUG, "writing <%02x> to port <%04x>", val, port););
	if ((port & B_ULA) == (P_ULA & B_ULA))
	{
		uns8 changes = val ^ Last0xFE;
		if ((changes & B_BORDER) != 0)
		{
			ScreenSetBorder(val & B_BORDER);
		}
#ifdef REAL_AUDIO
		if ((changes & (B_SPEAKER | B_MIC)) != 0)
		{
			SpeakerOutByte(val);
		}
#endif
		Last0xFE = val;
	}
	if ((port & B_SNDCHIP) == (P_SNDCONTROL & B_SNDCHIP))
	{
		if ((port & B_SNDCONTROL) == (P_SNDCONTROL & B_SNDCONTROL))
		{
			Last0xFFFD = val;
		}
		else
		{
			if (Last0xFFFD < NELEM(Psg))
			{
				Psg[Last0xFFFD] = val;
				if (Last0xFFFD == 14)
				{
					DEB(Msg(M_DEBUG, "writing AY register 14"););
				}
#ifdef REAL_AUDIO
				else if (Last0xFFFD < 14)
				{
					AyOutByte(Last0xFFFD, val);
				}
#endif
			}
			else
			{
					DEB(Msg(M_DEBUG, "writing AY register %d", Last0xFFFD););
			}
		}
	}
}
static uns8
didaktikInPort(uns16 port)
{
	uns8 res;
	if ((port & B_ULA) == (P_ULA & B_ULA))
	{
		res = 0xbf;
		INBYTE_KEYBOARD(port, res);
		res ^= TapeInGetBit();
		goto quit;
	}
	if ((port & B_SNDCHIP) == (P_SNDCONTROL & B_SNDCHIP))
	{
		if ((port & B_SNDCONTROL) == (P_SNDCONTROL & B_SNDCONTROL))
		{
			if (Last0xFFFD == 14)
			{
				DEB(Msg(M_DEBUG, "reading AY register 14"););
				res = Psg[7] & 0x40 ? (Psg[14] & 0x1f) | 0x20 : 0x3f;
			}
			else if (Last0xFFFD < 14)
			{
				res = (Psg[Last0xFFFD] & PsgMask[Last0xFFFD]);
			}
			else if (Last0xFFFD == 15)
			{
				res = Psg[7] & 0x80 ? Psg[15] : 0xff;
			}
			else
			{
				res = 0x00;
			}
		}
		else
		{
			res = 0xff;
		}
		goto quit;
	}
	res = 0xff;
  quit:
	if (port & B_ULA)
	DEB(Msg(M_DEBUG, "reading <%02x> from port <%04x>", res, port););
	return res;
}
#endif
