/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifdef XZX_KMOUSE
#ifndef lint
static char rcsid[] = "$Id: kmouse.c,v 1.14 2000/11/04 18:54:57 erik Rel $";
#endif
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <stdio.h>
#include "debug.h"
#include "resource.h"
#include "util.h"
//#include "screen.h"
#include "xwin.h"
#include "kmouse.h"
#define EVENT_MASK						(PointerMotionMask | ButtonPressMask | ButtonReleaseMask)
#define L_BUTTON						0x02
#define M_BUTTON						0x04
#define R_BUTTON						0x01
#define ZONE_OFFSET						(signed int)BORDER_PIXELS
#ifdef DEBUG
#define DEB(x)							{ if ((CFGGET(debug) & D_X11) != 0) { x } }
#else
#define DEB(x)
#endif
static void kmouseMove(Widget, XtPointer, XEvent *, Boolean *);
static void kmouseButton(Widget, XtPointer, XEvent *, Boolean *);
static void kmouseEnter(Widget, XtPointer, XEvent *, Boolean *);
uns8 KmouseX, KmouseY, KmouseButtons;
static Cursor blankCursor;
static int x, y;
void
KmouseInit(void)
{
	char bm;
	Pixmap pm;
	XColor c;
	XwinData *xdp = &XwinDatas;
	bm = 0;
	pm = XCreatePixmapFromBitmapData(xdp->display, XtWindow(xdp->emuWindow), &bm, 1, 1, 0, 0, 1);
	blankCursor = XCreatePixmapCursor(xdp->display, pm, pm, &c, &c, 0, 0);
	if (CFGGET(kmouse))
	{
		KmouseOn();
	}
}
void
KmouseOn(void)
{
	XwinData *xdp = &XwinDatas;
	XtAddEventHandler(xdp->emuWindow, PointerMotionMask, 0, (XtEventHandler)kmouseMove, (XtPointer)0);
	XtAddEventHandler(xdp->emuWindow, ButtonPressMask | ButtonReleaseMask, 0, (XtEventHandler)kmouseButton, (XtPointer)0);
	XtAddEventHandler(xdp->emuWindow, EnterWindowMask, 0, (XtEventHandler)kmouseEnter, (XtPointer)0);
	XDefineCursor(xdp->display, XtWindow(xdp->emuWindow), blankCursor);
	if (CFGGET(kmouseGrab))
	{
		(void)XGrabPointer(XwinDatas.display, XtWindow(xdp->emuWindow), True, PointerMotionMask, GrabModeAsync, GrabModeAsync,
						   XtWindow(xdp->emuWindow), None, CurrentTime);
		XWarpPointer(xdp->display, XtWindow(xdp->emuWindow), XtWindow(xdp->emuWindow), 0, 0, 0U, 0U, 0, 0);
		XSync(xdp->display, False);
	}
	KmouseX = x = 0;
	KmouseY = y = 0;
	KmouseButtons = 0xff;
}
void
KmouseOff(void)
{
	XwinData *xdp = &XwinDatas;
	XtRemoveEventHandler(xdp->emuWindow, PointerMotionMask, 0, (XtEventHandler)kmouseMove, (XtPointer)0);
	XtRemoveEventHandler(xdp->emuWindow, ButtonPressMask | ButtonReleaseMask, 0, (XtEventHandler)kmouseButton, (XtPointer)0);
	XtRemoveEventHandler(xdp->emuWindow, EnterWindowMask, 0, (XtEventHandler)kmouseEnter, (XtPointer)0);
	XUndefineCursor(xdp->display, XtWindow(xdp->emuWindow));
	XUngrabPointer(xdp->display, CurrentTime);
}
static void
kmouseMove(Widget w, XtPointer client, XEvent *xev, Boolean *cont)
{
	DEB(Msg(M_DEBUG, "KMOUSE: x %d, y %d", xev->xmotion.x, xev->xmotion.y););
	if (x != xev->xmotion.x || y != xev->xmotion.y)
	{
		XwinData *xdp = &XwinDatas;
		KmouseX -= (x - xev->xmotion.x) / CFGGET(scale);
		KmouseY += (y - xev->xmotion.y) / CFGGET(scale);
		if (CFGGET(kmouseGrab)
			&& (x < ZONE_OFFSET || y < ZONE_OFFSET
				|| x > ((int)(xdp->hsize) - ZONE_OFFSET) || y > ((int)(xdp->vsize) - ZONE_OFFSET)))
		{
			x = xdp->hsize >> 1;
			y = xdp->vsize >> 1;
			XWarpPointer(xdp->display, XtWindow(xdp->emuWindow), XtWindow(xdp->emuWindow), 0, 0, 0U, 0U, x, y);
		}
		else
		{
			x = xev->xmotion.x;
			y = xev->xmotion.y;
		}
	}
}
static void
kmouseButton(Widget w, XtPointer client, XEvent *xev, Boolean *cont)
{
	uns8 mask;
	switch (xev->xbutton.button)
	{
		case Button1:
			mask = L_BUTTON;
			break;
		case Button2:
			mask = M_BUTTON;
			break;
		case Button3:
			mask = R_BUTTON;
			break;
		default:
			mask = 0;
			break;
	}
	DEB(Msg(M_DEBUG, "KMOUSE: %s %x", xev->type == ButtonPress ? "press" : "release", mask););
	if (xev->type == ButtonPress)
	{
		KmouseButtons &= ~mask;
	}
	else
	{
		KmouseButtons |= mask;
	}
}
static void
kmouseEnter(Widget w, XtPointer client, XEvent *xev, Boolean *cont)
{
	if (CFGGET(kmouseGrab))
	{
		(void)XtGrabPointer(w, True, PointerMotionMask, GrabModeAsync, GrabModeAsync, XtWindow(w), None, CurrentTime);
		XSync(XwinDatas.display, False);
	}
}
#endif
