/********************************************************************************
 * Copyright (c) Des Herriott 1993, 1994
 *               Erik Kunze   1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation
 * for any purpose is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that the name
 * of the copyright holder not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.  The
 * copyright holder makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: Des Herriott
 *          Erik Kunze
 *******************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: main.c,v 4.34 2000/11/26 18:53:56 erik Rel $";
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <time.h>
#include "resource.h"
#include "mem.h"
#include "util.h"
#ifdef AUDIO
#include "audio.h"
#endif
#include "emul.h"
#ifdef JOY
#include "joystick.h"
#endif
#include "machine.h"
#ifdef PRINTER
#include "printer.h"
#endif
#include "screen.h"
#include "snapshot.h"
#include "tape.h"
#include "ui.h"
#include "xwin.h"
#include "main.h"
static int initEmulator(int, char **);
static void quitEmulator(void);
static void catchSignals(int);
char *ProgName;
const char *Version;
time_t StartupTime;
int StdinFileType, StdoutFileType;
int
main(int argc, char **argv)
{
	char *s;
	if ((s = strrchr(argv[0], '/')) != (char *)0)
	{
		ProgName = ++s;
	}
	else
	{
		ProgName = argv[0];
	}
#ifdef REGISTERED
	Version = "XZX " VERSION "R";
#else
	Version = "XZX " VERSION;
#endif
	StartupTime = time((time_t *)0);
	if (EXPIRATION && ((time_t)EXPIRATION < StartupTime))
	{
		(void)fprintf(stderr,
					  "Sorry, your version has expired! Get a new one.\n");
		return 1;
	}
	if (!initEmulator(argc, argv))
	{
		EmulLoop();
	}
	Quit(1);
	return 0;
}
static int
initEmulator(int ac, char **av)
{
	struct sigaction sigact;
	unsigned int i;
	sigact.sa_handler = SIG_IGN;
	(void)sigemptyset(&sigact.sa_mask);
#ifdef SA_RESTART
	sigact.sa_flags = SA_RESTART;
#else
	sigact.sa_flags = 0;
#endif
	(void)sigaction(SIGINT,   &sigact, (struct sigaction *)0);
	(void)sigaction(SIGHUP,   &sigact, (struct sigaction *)0);
	(void)sigaction(SIGTERM,  &sigact, (struct sigaction *)0);
	(void)sigaction(SIGQUIT,  &sigact, (struct sigaction *)0);
#ifdef SIG_XCPU
	(void)sigaction(SIG_XCPU, &sigact, (struct sigaction *)0);
#endif
	(void)sigaction(SIGPIPE,  &sigact, (struct sigaction *)0);
	sigact.sa_handler = catchSignals;
	(void)sigaction(SIGSEGV,  &sigact, (struct sigaction *)0);
	XwinInit(&ac, av);
	UiInit(!av[1]);
	SetBlocking(STDIN_FILENO, 1);
	StdinFileType = ClassifyDescriptor(STDIN_FILENO);
	SetBlocking(STDOUT_FILENO, 1);
	StdoutFileType = ClassifyDescriptor(STDOUT_FILENO);
	OnQuit(quitEmulator);
	MachineInit();
#ifdef REAL_AUDIO
	AudioInit();
#endif
	TapeInit();
#ifdef JOY
	JoyInit();
#endif
#ifdef PRINTER
	PrinterInit();
#endif
	Machine->reset();
	if (!av[1] || SnapshotRead(av[1]) == -1)
	{
		srand((unsigned int)getpid());
		for (i = DFILE; i < (DFILE + PIXEL_LENGTH + ATTR_LENGTH); i++)
		{
			MEM(i) = (uns8)rand();
		}
	}
	ScreenRefresh();
	(void)sigaction(SIGINT,  &sigact, (struct sigaction *)0);
	(void)sigaction(SIGHUP,  &sigact, (struct sigaction *)0);
	(void)sigaction(SIGTERM, &sigact, (struct sigaction *)0);
	(void)sigaction(SIGQUIT, &sigact, (struct sigaction *)0);
	(void)sigaction(SIGSEGV, &sigact, (struct sigaction *)0);
	return 0;
}
static void
quitEmulator(void)
{
	SetBlocking(STDIN_FILENO, 0);
	SetBlocking(STDOUT_FILENO, 0);
}
static void
catchSignals(int signo)
{
	static int core = 0;
	if (signo == SIGSEGV)
	{
		if (++core > 1)
		{
			exit(3);
		}
		Msg(M_ERR, "signal handler caught signal SIGSEGV");
		Msg(M_ERR, "send a bug report to xzx-bug@fantasy.muc.de");
		if (CFGGET(core))
		{
			abort();
		}
	}
	Quit(2);
}
