/*********************************************************************************************************************************
 * Copyright (c) Des Herriott 1993 - 1994
 *               Erik Kunze   1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: Des Herriott
 *          Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: snapshot.c,v 4.43 2000/11/04 18:28:11 erik Rel $";
#endif
#include <stdlib.h>
#include <string.h>
#include "cpu.h"
#include "debug.h"
#include "resource.h"
#include "mem.h"
#include "machine.h"
#include "util.h"
#include "dat.h"
#include "screen.h"
#include "slt.h"
#include "sna.h"
#include "z80.h"
#include "snapshot.h"
#ifdef DEBUG
#define DEB(x)							{ if ((CFGGET(debug) & D_SNAP) != 0) { x } }
#else
#define DEB(x)
#endif
static char *snapshotLastFname = (char *)0;
int
SnapshotRead(char *fname)
{
	int ret = -1;
	DEB(Msg(M_DEBUG, "last snaphot was <%s>", GetBaseName(snapshotLastFname));
		Msg(M_DEBUG, "now loading <%s>", GetBaseName(fname)););
	Machine->reset();
	switch (GetFileType(fname))
	{
		case FT_UNKNOWN:
		case FT_SNA:
			ret = SnaRead(fname);
			break;
		case FT_Z80:
			ret = Z80Read(fname);
			break;
		case FT_SLT:
			ret = SltRead(fname);
			break;
	}
	ScreenMarkDirtyAll();
	if (snapshotLastFname)
	{
		free(snapshotLastFname);
	}
	snapshotLastFname = ret == 0 ? Strdup(fname, "SnapshotRead") : (char *)0;
	return ret;
}
void
SnapshotReadLevel(void)
{
	int ret;
	if (snapshotLastFname == (char *)0)
	{
		Msg(M_WARN, "no previous snapshot name - multiload ignored");
		return;
	}
	if (READ_ONLY(HL))
	{
		Msg(M_WARN, "attempt to read level into ROM - ignored");
		return;
	}
	if (GetFileType(snapshotLastFname) == FT_SLT)
	{
		ret = SltReadLevel(snapshotLastFname);
	}
	else
	{
		ret = DatReadLevel(snapshotLastFname);
	}
	if (ret < 0)
	{
		CCF();
	}
}
int
SnapshotWrite(char *fname)
{
	int type, ret = -1;
	type = GetFileType(fname);
	if (CFGGET(machine) > SP_48_3 && type != FT_SLT)
	{
		type = FT_Z80;
	}
	switch(type)
	{
		case FT_SNA:
			ret = SnaWrite(fname);
			break;
		case FT_UNKNOWN:
		case FT_Z80:
			ret = Z80Write(fname);
			break;
		case FT_SLT:
			if (snapshotLastFname == (char *)0 || GetFileType(snapshotLastFname) != FT_SLT)
			{
				Msg(M_WARN, "attempt to write SLT snapshots in non-SLT mode - ignored");
			}
			else if (strcmp(snapshotLastFname, fname) == 0)
			{
				Msg(M_WARN, "attempt to overwrite current SLT - ignored");
			}
			else
			{
				ret = SltWrite(fname, snapshotLastFname);
			}
			break;
	}
	return ret;
}
