/********************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation
 * for any purpose is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that the name
 * of the copyright holder not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.  The
 * copyright holder makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 *******************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#if defined(AIX) && !defined(NAS_AUDIO) && !defined(OSS_AUDIO)
#ifndef lint
static char rcsid[] = "$Id: aix.c,v 1.6 2000/11/07 05:14:32 erik Rel $";
#endif
#include <sys/types.h>
#include <signal.h>
#include <UMS/UMSAudioDevice.h>
#include <UMS/UMSBAUDDevice.h>
#include "debug.h"
#include "resource.h"
#include "util.h"
#include "audio.h"
#ifdef DEBUG
#define DEB(x)			{ if (CFGGET(debug) & D_AUDIO) { x } }
#else
#define DEB(x)
#endif
static Environment *ev;
static UMSBAUDDevice audioDevice = (UMSBAUDDevice)0;
static UMSAudioTypes_Buffer buffer;
static sigset_t blockingSet;
int
OsdOpenAudioDevice(char *device, int *sample, int *stereo)
{
	UMSAudioDevice_ReturnCode rc;
	long s, g;
	int i;
	Msg(M_INFO, "*******************************************************");
	Msg(M_INFO, "This audio driver is experimental, because I don't have");
	Msg(M_INFO, "access to any system running AIX. Please get in contact");
	Msg(M_INFO, "with me so I can improve the driver.");
	Msg(M_INFO, "*******************************************************");
	ev = somGetGlobalEnvironment();
	audioDevice = UMSBAUDDeviceNew();
	(void)UMSAudioDevice_open(audioDevice, ev, "/dev/paud0", "PLAY",
							  UMSAudioDevice_BlockingIO);
	if (audioDevice == (UMSBAUDDevice)0)
	{
		Msg(M_ERR, "couldn't create audio device object");
		return -1;
	}
	if (UMSAudioDevice_set_volume(audioDevice, ev, 100) < 0
		|| UMSAudioDevice_set_balance(audioDevice, ev, 0) < 0
		|| UMSAudioDevice_set_time_format(audioDevice, ev,
										  UMSAudioTypes_Msecs) < 0
		|| UMSAudioDevice_set_byte_order(audioDevice, ev, "LSB"))
	{
		Msg(M_ERR, "couldn't set audio-device parameters");
		goto error;
	}
	if (UMSAudioDevice_set_bits_per_sample(audioDevice, ev, 8) < 0
		|| UMSAudioDevice_set_audio_format_type(audioDevice, ev, "PCM") < 0
		|| UMSAudioDevice_set_number_format(audioDevice, ev,
											"TWOS_COMPLEMENT") < 0)
	{
		Msg(M_ERR, "couldn't set sample resolution to 8 bit");
		goto error;
	}
	if (UMSAudioDevice_set_sample_rate(audioDevice, ev, *sample, &s) < 0)
	{
		Msg(M_ERR, "couldn't set sample rate to %d Hz", *sample);
		goto error;
	}
	*sample = (int)s;
	i = 2;
	if (UMSAudioDevice_set_number_of_channels(audioDevice, ev, i) < 0)
	{
		Msg(M_WARN, "couldn't set stereo output, forcing mono");
		i = 1;
		if (UMSAudioDevice_set_number_of_channels(audioDevice, ev, i) < 0)
		{
			Msg(M_ERR, "couldn't set mono output");
			goto error;
		}
	}
	*stereo = i == 2;
	buffer._maximum = (long)OsdGetAudioFreespace();
	buffer._buffer = (char *)0;
	buffer._length = 0;
	g = 100;
	if (UMSAudioDevice_enable_output(audioDevice, ev, "LINE_OUT", &g, &g) < 0
		|| UMSAudioDevice_initialize(audioDevice, ev) < 0
		|| UMSAudioDevice_start(audioDevice, ev) < 0)
	{
		Msg(M_ERR, "couldn't start audio device");
		goto error;
	}
	sigemptyset(&blockingSet);
	sigaddset(&blockingSet, SIGALRM);
	return 0;
  error:
	UMSAudioDevice_stop(audioDevice, ev);
	UMSAudioDevice_close(audioDevice, ev);
	_somFree(audioDevice);
	audioDevice = (UMSBAUDDevice)0;
	return -1;
}
void
OsdCloseAudioDevice(void)
{
	if (audioDevice != (UMSBAUDDevice)0)
	{
		(void)UMSAudioDevice_play_remaining_data(audioDevice, ev, TRUE);
		(void)UMSAudioDevice_stop(audioDevice, ev);
		(void)UMSAudioDevice_close(audioDevice, ev);
		_somFree(audioDevice);
		audioDevice = (UMSBAUDDevice)-1;
	}
}
int
OsdDrainAudioDevice(void)
{
	return ((int)UMSAudioDevice_play_remaining_data(audioDevice, ev, TRUE));
}
size_t
OsdGetAudioFreespace(void)
{
	int i = -1;
	return ((size_t)(UMSAudioDevice_write_buff_remain(audioDevice, ev, &i) < 0 ?
					 -1 : i));
}
size_t
OsdPlayAudioBuffer(unsigned char *buf, size_t len)
{
	sigset_t oldSet;
	long written;
	buffer._buffer = buf;
	buffer._length = (long)len;
	(void)sigprocmask(SIG_BLOCK, &blockingSet, &oldSet);
	if (UMSAudioDevice_write(audioDevice, ev, &buffer, len, &written) < 0)
	{
		Msg(M_PERR, "couldn't write to audio device");
	}
	if (!sigismember(&oldSet, SIGALRM))
	{
		(void)sigprocmask(SIG_UNBLOCK, &blockingSet, (sigset_t *)0);
	}
	return (size_t)written;
}
#endif
