/********************************************************************************
 * Copyright (c) Des Herriott 2000
 *               Erik Kunze   1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation
 * for any purpose is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that the name
 * of the copyright holder not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.  The
 * copyright holder makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Des Herriott
 *         Erik Kunze
 *******************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifdef ALSA_AUDIO
#ifndef lint
static char rcsid[] = "$Id: alsa.c,v 1.2 2000/11/07 05:14:32 erik Rel $";
#endif
#include <sys/types.h>
#include <signal.h>
#include <sys/asoundlib.h>
#include "debug.h"
#include "resource.h"
#include "util.h"
#include "audio.h"
#ifdef DEBUG
#define DEB(x)			{ if (CFGGET(debug) & D_AUDIO) { x } }
#else
#define DEB(x)
#endif
static snd_pcm_t *pcmHandle = (snd_pcm_t *)0;
static sigset_t blockingSet;
int
OsdOpenAudioDevice(char *device, int *sample, int *stereo)
{
	snd_pcm_channel_info_t info;
	snd_pcm_channel_params_t params;
	int card = 0, dev = 0;
	Msg(M_INFO, "********************************************************");
	Msg(M_INFO, "This audio driver is experimental, because I don't have");
	Msg(M_INFO, "access to any system running ALSA. Please get in contact");
	Msg(M_INFO, "with me so I can improve the driver.");
	Msg(M_INFO, "********************************************************");
	if (snd_pcm_open(&pcmHandle, card, dev,
					 SND_PCM_OPEN_PLAYBACK | SND_PCM_OPEN_NONBLOCK) < 0)
	{
		Msg(M_PERR, "couldn't open device %d, card %d for writing", dev, card);
		return -1;
	}
	if (snd_pcm_channel_info(pcmHandle, &info) < 0)
	{
		Msg(M_PERR, "couldn't get audio settings");
		goto error;
	}
	DEB(Msg(M_DEBUG, "AU: min sample rate  %u", info.min_rate);
		Msg(M_DEBUG, "AU: max sample rate  %u", info.max_rate);
		Msg(M_DEBUG, "AU: channels         %d", info.max_voices););
	if (*sample > info.max_rate)
	{
		*sample = info.max_rate;
	}
	else if (*sample < info.min_rate)
	{
		*sample = info.min_rate;
	}
	memset(&params, 0, sizeof(params));
	params.channel = SND_PCM_CHANNEL_PLAYBACK;
	params.mode = SND_PCM_MODE_STREAM;
	params.start_mode = SND_PCM_START_DATA;
	params.stop_mode = SND_PCM_STOP_ROLLOVER;
	params.format.format = SND_PCM_SFMT_U8;
	params.format.rate = *sample;
	params.format.voices = info.max_voices > 1 ? 2 : 1;
	params.format.interleave = 1;
	if (snd_pcm_channel_params(pcmHandle, &params) < 0)
	{
		Msg(M_ERR, "couldn't set 8 bit, %d Hz, %s", params.format.rate,
			params.format.voices == 2 ? "stereo" : "mono");
		goto error;
	}
	*stereo = params.format.voices == 2;
	sigemptyset(&blockingSet);
	sigaddset(&blockingSet, SIGALRM);
	return 0;
  error:
	(void)snd_pcm_close(pcmHandle);
	pcmHandle = (snd_pcm_t *)0;
	return -1;
}
void
OsdCloseAudioDevice(void)
{
	if (pcmHandle != (snd_pcm_t *)0)
	{
		(void)snd_pcm_playback_flush(pcmHandle);
		(void)snd_pcm_close(pcmHandle);
		pcmHandle = (snd_pcm_t *)0;
	}
}
int
OsdDrainAudioDevice(void)
{
	return (snd_pcm_playback_drain(pcmHandle));
}
size_t
OsdGetAudioFreespace(void)
{
	snd_pcm_channel_status_t status;
	return ((size_t)(snd_pcm_channel_status(pcmHandle, &status) < 0 ?
					 -1 : status.free));
}
size_t
OsdPlayAudioBuffer(unsigned char *buf, size_t len)
{
	sigset_t oldSet;
	size_t written;
	(void)sigprocmask(SIG_BLOCK, &blockingSet, &oldSet);
	if ((written = snd_pcm_write(pcmHandle, buf, len)) < 0)
	{
		Msg(M_PERR, "couldn't write to audio device");
	}
	if (!sigismember(&oldSet, SIGALRM))
	{
		(void)sigprocmask(SIG_UNBLOCK, &blockingSet, (sigset_t *)0);
	}
	return written;
}
#endif
