/********************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation
 * for any purpose is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that the name
 * of the copyright holder not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.  The
 * copyright holder makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 *******************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifdef NAS_AUDIO
#ifndef lint
static char rcsid[] = "$Id: nas.c,v 4.9 2000/11/07 05:14:32 erik Rel $";
#endif
#include <sys/types.h>
#include <signal.h>
#include <audio/audiolib.h>
#include "debug.h"
#include "resource.h"
#include "util.h"
#include "audio.h"
#ifdef DEBUG
#define DEB(x)			{ if (CFGGET(debug) & D_AUDIO) { x } }
#else
#define DEB(x)
#endif
static AuBool nasIoError(AuServer *);
static AuServer *nasServer = (AuServer *)0;
static AuDeviceID nasDevice = AuNone;
static AuFlowID nasChannel;
static int nasQueueSize;
static sigset_t blockingSet;
int
OsdOpenAudioDevice(char *device, int *sample, int *stereo)
{
	AuElement elements[2];
	AuElementAction actions[1];
	AuStatus status;
	char msg[80];
	int tracks, i;
	if (!(nasServer = AuOpenServer((char *)0, 0, (char *)0, 0, (char *)0,
								   (char **)0)))
	{
		Msg(M_ERR, "couldn't connect to audio server");
		return -1;
	}
	for  (tracks = 2; tracks; tracks--)
	{
		for (i = 0; i < AuServerNumDevices(nasServer); i++)
		{
			if ((AuDeviceKind(AuServerDevice(nasServer, i)) ==
				 AuComponentKindPhysicalOutput)
				&& AuDeviceNumTracks(AuServerDevice(nasServer, i)) == tracks)
			{
				nasDevice = AuDeviceIdentifier(AuServerDevice(nasServer, i));
				goto found;
			}
		}
	}
  found:
	if (nasDevice == AuNone)
	{
		Msg(M_ERR, "couldn't find output device");
		goto error;
	}
	if (!(nasChannel = AuCreateFlow(nasServer, (AuStatus *)0)))
	{
		Msg(M_ERR, "couldn't create flow");
		goto error;
	}
	DEB(Msg(M_DEBUG, "AU: server           %s",
			AuServerString(nasServer));
		Msg(M_DEBUG, "AU: min sample rate  %u",
			AuServerMinSampleRate(nasServer));
		Msg(M_DEBUG, "AU: max sample rate  %u",
			AuServerMaxSampleRate(nasServer));
		Msg(M_DEBUG, "AU: channels         %u",
			tracks););
	if (*sample > AuServerMaxSampleRate(nasServer))
	{
		*sample = AuServerMaxSampleRate(nasServer);
	}
	else if (*sample < AuServerMinSampleRate(nasServer))
	{
		*sample = AuServerMinSampleRate(nasServer);
	}
	nasQueueSize = *sample * tracks;
	(void)AuSetIOErrorHandler(nasServer, nasIoError);
	AuMakeChangeStateAction(&actions[0],
							AuStateStop, AuStateStart, AuReasonEOF,
							nasChannel, AuElementAll, AuStateStart);
	AuMakeElementImportClient(&elements[0],
							  (unsigned short)*sample, AuFormatLinearUnsigned8,
							  (unsigned char)tracks,
							  AuTrue,
							  (AuUint32)nasQueueSize, (AuUint32)0,
							  1, actions);
	AuMakeElementExportDevice(&elements[1], 0,
							  nasDevice,
							  (unsigned short)*sample, AuUnlimitedSamples,
							  0, (AuElementAction *)0);
	AuSetElements(nasServer, nasChannel, AuTrue, 2, elements, &status);
	if (status != AuSuccess)
	{
		AuGetErrorText(nasServer, status, msg, sizeof(msg));
		Msg(M_ERR, "couldn't config flow: %s", msg);
		goto error;
	}
	AuStartFlow(nasServer, nasChannel, &status);
	if (status != AuSuccess)
	{
		AuGetErrorText(nasServer, status, msg, sizeof(msg));
		Msg(M_ERR, "couldn't start flow: %s", msg);
		goto error;
	}
	*stereo = tracks == 2;
	sigemptyset(&blockingSet);
	sigaddset(&blockingSet, SIGALRM);
	return 0;
  error:
	AuCloseServer(nasServer);
	nasServer = (AuServer *)0;
	return -1;
}
void
OsdCloseAudioDevice(void)
{
	if (nasServer != (AuServer *)0)
	{
		AuStopFlow(nasServer, nasChannel, (AuStatus *)0);
		AuCloseServer(nasServer);
		nasServer = (AuServer *)0;
	}
}
int
OsdDrainAudioDevice(void)
{
	AuFlush(nasServer);
	return 0;
}
size_t
OsdGetAudioFreespace(void)
{
	return (size_t)nasQueueSize;
}
size_t
OsdPlayAudioBuffer(unsigned char *buf, size_t len)
{
	sigset_t oldSet;
	(void)sigprocmask(SIG_BLOCK, &blockingSet, &oldSet);
	AuWriteElement(nasServer, nasChannel, 0, (AuUint32)len, (AuPointer)buf,
				   AuFalse, (AuStatus *)0);
	if (!sigismember(&oldSet, SIGALRM))
	{
		(void)sigprocmask(SIG_UNBLOCK, &blockingSet, (sigset_t *)0);
	}
	return len;
}
static AuBool
nasIoError(AuServer *server)
{
	Msg(M_FATAL, "non recoverable error in NAS audio system");
	return AuTrue;
}
#endif
