/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: resource.c,v 4.62 2000/11/26 18:48:28 erik Rel $";
#endif
#include <Xm/Xm.h>
#include <X11/keysym.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
#include "debug.h"
#include "machine.h"
#include "util.h"
#ifdef AUDIO
#include "audio.h"
#endif
#ifdef XZX_DIDAKTIK
#include "didaktik.h"
#endif
#include "fdc.h"
#ifdef XZX_IF1
#include "if1.h"
#endif
#include "main.h"
#ifdef XZX_MGT
#include "mgt.h"
#endif
#ifdef XZX_TRDOS
#include "trdos.h"
#endif
#include "xwin.h"
#include "resource.h"
#define RESOURCE_FILE					".xzxrc"
#define RES(a,b,c,d)					{ (char *)a,(char *)b,c,d },
#define RES2(a,b,c,d,e,f,g)				{ (String)a,(String)b,c,sizeof(d),XtOffsetOf(ResourceData,e),f,(XtPointer)g },
#define RES3(a,b,c,d,e,f,g)				{ (String)a,(String)b,c,sizeof(d),(size_t)e,f,(XtPointer)g },
#define BOOL(x)							(x > 0 ? "true" : "false")
#define KEY(x)							XKeysymToString(x)
#define XtNull 							(XtPointer)0
enum { KM_JOY_UP, KM_JOY_DOWN, KM_JOY_LEFT, KM_JOY_RIGHT, KM_JOY_FIRE, NUM_KEYMAPPINGS };
static void resourcePrintCopyright(void);
static void resourcePrintVersion(void);
#ifdef DEBUG
static void resourcePrint(void);
#endif
ResourceData ResourceDatas;
XrmOptionDescRec ResourceOptions[] = {
	RES("-libDir",         "*libDir",        XrmoptionSepArg, XtNull)
	RES("-infoDir",        "*infoDir",       XrmoptionSepArg, XtNull)
	RES("-pokeDir",        "*pokeDir",       XrmoptionSepArg, XtNull)
	RES("-quiet",          "*quiet",         XrmoptionSepArg, XtNull)
	RES("-core",           "*core",          XrmoptionSepArg, XtNull)
	RES("-monocrome",      "*monocrome",     XrmoptionSepArg, XtNull)
	RES("-private",        "*private",       XrmoptionSepArg, XtNull)
	RES("-xsync",          "*xsync",         XrmoptionSepArg, XtNull)
#ifdef DEBUG
	RES("-debug",          "*debug",         XrmoptionSepArg, XtNull)
#endif
	RES("-machine",        "*machine",       XrmoptionSepArg, XtNull)
	RES("-speed",          "*speed",         XrmoptionSepArg, XtNull)
	RES("-fastmode",       "*fastmode",      XrmoptionSepArg, XtNull)
	RES("-warn48",         "*warn48",        XrmoptionSepArg, XtNull)
	RES("-scale",          "*scale",         XrmoptionSepArg, XtNull)
	RES("-refresh",        "*refresh",       XrmoptionSepArg, XtNull)
	RES("-cache",          "*cache",         XrmoptionSepArg, XtNull)
#ifndef NO_ED_TRAPS
	RES("-tapeQuickload",  "*tapeQuickload", XrmoptionSepArg, XtNull)
#endif
	RES("-tapeFastload",   "*tapeFastload",  XrmoptionSepArg, XtNull)
	RES("-tapeAutostop",   "*tapeAutostop",  XrmoptionSepArg, XtNull)
#ifdef REAL_AUDIO
	RES("-tapeNoise",      "*tapeNoise",     XrmoptionSepArg, XtNull)
#endif
	RES("-joyUp",          "*joyUp",         XrmoptionSepArg, XtNull)
	RES("-joyDown",        "*joyDown",       XrmoptionSepArg, XtNull)
	RES("-joyLeft",        "*joyLeft",       XrmoptionSepArg, XtNull)
	RES("-joyRight",       "*joyRight",      XrmoptionSepArg, XtNull)
	RES("-joyFire",        "*joyFire",       XrmoptionSepArg, XtNull)
#ifdef JOY
	RES("-joy",            "*joy",           XrmoptionSepArg, XtNull)
	RES("-joyDev",         "*joyDev",        XrmoptionSepArg, XtNull)
	RES("-joyTolerance",   "*joyTolerance",  XrmoptionSepArg, XtNull)
	RES("-joyAutofire",    "*joyAutofire",   XrmoptionSepArg, XtNull)
	RES("-joyJuggle",      "*joyJuggle",     XrmoptionSepArg, XtNull)
#endif
#ifdef AUDIO
	RES("-audio",          "*audio",         XrmoptionSepArg, XtNull)
#ifdef REAL_AUDIO
#ifdef NAS_AUDIO
	RES("-audioServer",    "*audioServer",   XrmoptionSepArg, XtNull)
#else
	RES("-audioDev",       "*audioDev",      XrmoptionSepArg, XtNull)
#endif
	RES("-audioSample",    "*audioSample",   XrmoptionSepArg, XtNull)
	RES("-audioStereo",    "*audioStereo",   XrmoptionSepArg, XtNull)
#endif
#ifdef XBELL_AUDIO
	RES("-audioVolume",    "*audioVolume",   XrmoptionSepArg, XtNull)
#endif
#endif
	RES("-sp48Rom",        "*sp48Rom",       XrmoptionSepArg, XtNull)
	RES("-sp128Rom",       "*sp128Rom",      XrmoptionSepArg, XtNull)
	RES("-sp3Rom",         "*sp3Rom",        XrmoptionSepArg, XtNull)
	RES("-sp3DriveA",      "*sp3DriveA",     XrmoptionSepArg, XtNull)
	RES("-sp3DriveB",      "*sp3DribeB",     XrmoptionSepArg, XtNull)
#ifdef XZX_PENTAGON
	RES("-pentagonRom",    "*pentagonRom",   XrmoptionSepArg, XtNull)
#endif
#ifdef XZX_SCORPION
	RES("-scorpionRom",    "*scorpionRom",   XrmoptionSepArg, XtNull)
#endif
#ifdef XZX_DIDAKTIK
	RES("-didaktikRom",    "*didaktikRom",   XrmoptionSepArg, XtNull)
	RES("-didaktikDriveA", "*didaktikRom",   XrmoptionSepArg, XtNull)
	RES("-didaktikDriveB", "*didaktikRom",   XrmoptionSepArg, XtNull)
#endif
#ifdef XZX_IF1
	RES("-if1",            "*if1",           XrmoptionSepArg, XtNull)
	RES("-if1Rom",         "*if1Rom",        XrmoptionSepArg, XtNull)
	RES("-if1Crlf",        "*if1Crlf",       XrmoptionSepArg, XtNull)
	RES("-if1Strcr",       "*if1Strcr",      XrmoptionSepArg, XtNull)
	RES("-if1Drive1",      "*if1Drive1",     XrmoptionSepArg, XtNull)
#if IF1_DRIVES > 1
	RES("-if1Drive2",      "*if1Drive2",     XrmoptionSepArg, XtNull)
#endif
#if IF1_DRIVES > 2
	RES("-if1Drive3",      "*if1Drive3",     XrmoptionSepArg, XtNull)
#endif
#if IF1_DRIVES > 3
	RES("-if1Drive4",      "*if1Drive4",     XrmoptionSepArg, XtNull)
#endif
#if IF1_DRIVES > 4
	RES("-if1Drive5",      "*if1Drive5",     XrmoptionSepArg, XtNull)
#endif
#if IF1_DRIVES > 5
	RES("-if1Drive6",      "*if1Drive6",     XrmoptionSepArg, XtNull)
#endif
#if IF1_DRIVES > 6
	RES("-if1Drive7",      "*if1Drive7",     XrmoptionSepArg, XtNull)
#endif
#if IF1_DRIVES > 7
	RES("-if1Drive8",      "*if1Drive8",     XrmoptionSepArg, XtNull)
#endif
#endif
#ifdef XZX_MF
	RES("-mf",             "*mf",            XrmoptionSepArg, XtNull)
	RES("-mf1Rom",         "*mf1Rom",        XrmoptionSepArg, XtNull)
	RES("-mf128Rom",       "*mf128Rom",      XrmoptionSepArg, XtNull)
	RES("-mf3Rom",         "*mf3Rom",        XrmoptionSepArg, XtNull)
#endif
#ifdef XZX_TRDOS
	RES("-trdos",          "*trdos",         XrmoptionSepArg, XtNull)
	RES("-trdosRom",       "*trdosRom",      XrmoptionSepArg, XtNull)
	RES("-trdosDriveA",    "*trdosDriveA",   XrmoptionSepArg, XtNull)
	RES("-trdosDriveB",    "*trdosDriveB",   XrmoptionSepArg, XtNull)
	RES("-trdosDriveC",    "*trdosDriveC",   XrmoptionSepArg, XtNull)
	RES("-trdosDriveD",    "*trdosDriveD",   XrmoptionSepArg, XtNull)
#endif
#ifdef XZX_MGT
	RES("-mgt",            "*mgt",           XrmoptionSepArg, XtNull)
	RES("-mgtRom",         "*mgtRom",        XrmoptionSepArg, XtNull)
	RES("-mgtDriveA",      "*mgtDriveA",     XrmoptionSepArg, XtNull)
	RES("-mgtDriveB",      "*mgtDriveB",     XrmoptionSepArg, XtNull)
#endif
#ifdef XZX_KMOUSE
	RES("-kmouse",         "*kmouse",        XrmoptionSepArg, XtNull)
	RES("-kmouseGrab",     "*kmouseGrab",    XrmoptionSepArg, XtNull)
#endif
};
Cardinal ResourceOptionsCount = XtNumber(ResourceOptions);
static XtResource resourceDefaults[] = {
	RES2("libDir",         "LibDir",         XmRString,  XPointer,     libDir,           XmRString,    "."           )
	RES2("infoDir",        "InfoDir",        XmRString,  XPointer,     infoDir,          XmRString,    "."           )
	RES2("pokeDir",        "PokeDir",        XmRString,  XPointer,     pokeDir,          XmRString,    "."           )
	RES2("quiet",          "Quiet",          XmRBoolean, Boolean,      quiet,            XmRImmediate, FALSE         )
	RES2("core",           "Core",           XmRBoolean, Boolean,      core,             XmRImmediate, FALSE         )
	RES2("monocrome",      "Monocrome",      XmRBoolean, Boolean,      monocrome,        XmRImmediate, FALSE         )
	RES2("private",        "Private",        XmRBoolean, Boolean,      private,          XmRImmediate, FALSE         )
	RES2("xsync",          "Xsync",          XmRBoolean, Boolean,      xsync,            XmRImmediate, FALSE         )
#ifdef DEBUG
	RES2("debug",          "Debug",          XmRInt,     unsigned int, debug,            XmRImmediate, 0             )
#endif
	RES2("speed",          "Speed",          XmRInt,     int,          speed,            XmRImmediate, 100           )
	RES2("fastmode",       "Fastmode",       XmRBoolean, Boolean,      fastmode,         XmRImmediate, FALSE         )
	RES2("warn48",         "Warn48",         XmRBoolean, Boolean,      warn48,           XmRImmediate, FALSE         )
	RES2("scale",          "Scale",          XmRInt,     int,          scale,            XmRImmediate, 1             )
	RES2("refresh",        "Refresh",        XmRInt,     int,          refresh,          XmRImmediate, 3             )
	RES2("cache",          "Cache",          XmRBoolean, Boolean,      cache,            XmRImmediate, TRUE          )
#ifndef NO_ED_TRAPS
	RES2("tapeQuickload",  "TapeQuickload",  XmRBoolean, Boolean,      tapeQuickload,    XmRImmediate, TRUE          )
#endif
	RES2("tapeFastload",   "TapeFastload",   XmRBoolean, Boolean,      tapeFastload,     XmRImmediate, FALSE         )
	RES2("tapeAutostop",   "TapeAutostop",   XmRBoolean, Boolean,      tapeAutostop,     XmRImmediate, FALSE         )
#ifdef REAL_AUDIO
	RES2("tapeNoise",      "TapeNoise",      XmRBoolean, Boolean,      tapeNoise,        XmRImmediate, FALSE         )
#endif
#ifdef JOY
	RES2("joy",            "Joy",            XmRBoolean, Boolean,      joy,              XmRImmediate, FALSE         )
	RES2("joyDev",         "JoyDev",         XmRString,  XPointer,     joyDev,           XmRString,    "/dev/js0"    )
	RES2("joyTolerance",   "JoyTolerance",   XmRInt,     int,          joyTolerance,     XmRImmediate, 20            )
	RES2("joyAutofire",    "JoyAutofire",    XmRInt,     int,          joyAutofire,      XmRImmediate, 25            )
	RES2("joyJuggle",      "JoyJuggle",      XmRInt,     int,          joyJuggle,        XmRImmediate, 25            )
#endif
#ifdef AUDIO
	RES2("audio",          "Audio",          XmRBoolean, Boolean,      audio,            XmRImmediate, FALSE         )
#ifdef REAL_AUDIO
#if defined(NAS_AUDIO)
	RES2("audioServer",    "AudioServer",    XmRString,  XPointer,     audioDev,         XmRString,    0             )
#elif defined(OSS_AUDIO)
	RES2("audioDev",       "AudioDev",       XmRString,  XPointer,     audioDev,         XmRString,    "/dev/dsp"    )
#else
	RES2("audioDev",       "AudioDev",       XmRString,  XPointer,     audioDev,         XmRString,    "/dev/audio"  )
#endif
	RES2("audioSample",    "AudioSample",    XmRInt,     int,          audioSample,      XmRImmediate, 8000          )
	RES2("audioStereo",    "AudioStereo",    XmRInt,     int,          audioStereo,      XmRImmediate, AY_MONO       )
#endif
#ifdef XBELL_AUDIO
	RES2("audioVolume",    "AudioVolume",    XmRInt,     unsigned int, audioVolume,      XmRImmediate, 50            )
#endif
#endif
	RES2("sp48Rom",        "Sp48Rom",        XmRString,  XPointer,     sp48Rom,          XmRString,    "spectrum.rom")
	RES2("sp128Rom",       "Sp128Rom",       XmRString,  XPointer,     sp128Rom,         XmRString,    "128.rom"     )
	RES2("sp3Rom",         "Sp3Rom",         XmRString,  XPointer,     sp3Rom,           XmRString,    "plus3.rom"   )
	RES2("sp3DriveA",      "Sp3DriveA",      XmRString,  XPointer,     sp3Disks[0],      XmRString,    "diska.dsk"   )
	RES2("sp3DriveB",      "Sp3DriveB",      XmRString,  XPointer,     sp3Disks[1],      XmRString,    "diskb.dsk"   )
#ifdef XZX_PENTAGON
	RES2("pentagonRom",    "PentagonRom",    XmRString,  XPointer,     pentagonRom,      XmRString,    "pentagon.rom")
#endif
#ifdef XZX_SCORPION
	RES2("scorpionRom",    "ScorpionRom",    XmRString,  XPointer,     scorpionRom,      XmRString,    "scorpion.rom")
#endif
#ifdef XZX_DIDAKTIK
	RES2("didaktikRom",    "DidaktikRom",    XmRString,  XPointer,     didaktikRom,      XmRString,    "didaktik.rom")
	RES2("didaktikDriveA", "DidaktikDriveA", XmRString,  XPointer,     didaktikDisks[0], XmRString,    "diska.d80"   )
	RES2("didaktikDriveB", "DidaktikDriveB", XmRString,  XPointer,     didaktikDisks[1], XmRString,    "diskb.d80"   )
#endif
#ifdef XZX_IF1
	RES2("if1",            "If1",            XmRBoolean, Boolean,      if1,              XmRImmediate, FALSE         )
	RES2("if1Rom",         "If1Rom",         XmRString,  XPointer,     if1Rom,           XmRString,    "if1-v2.rom"  )
	RES2("if1Crlf",        "If1Crlf",        XmRBoolean, Boolean,      if1Crlf,          XmRImmediate, FALSE         )
	RES2("if1Strcr",       "If1Strcr",       XmRBoolean, Boolean,      if1Strcr,         XmRImmediate, FALSE         )
	RES2("if1Drive1",      "If1Drive1",      XmRString,  XPointer,     if1Carts[0],      XmRString,    "cart1.mdr"   )
#if IF1_DRIVES > 1
	RES2("if1Drive2",      "If1Drive2",      XmRString,  XPointer,     if1Carts[1],      XmRString,    "cart2.mdr"   )
#endif
#if IF1_DRIVES > 2
	RES2("if1Drive3",      "If1Drive3",      XmRString,  XPointer,     if1Carts[2],      XmRString,    "cart3.mdr"   )
#endif
#if IF1_DRIVES > 3
	RES2("if1Drive4",      "If1Drive4",      XmRString,  XPointer,     if1Carts[3],      XmRString,    "cart4.mdr"   )
#endif
#if IF1_DRIVES > 4
	RES2("if1Drive5",      "If1Drive5",      XmRString,  XPointer,     if1Carts[4],      XmRString,    "cart5.mdr"   )
#endif
#if IF1_DRIVES > 5
	RES2("if1Drive6",      "If1Drive6",      XmRString,  XPointer,     if1Carts[5],      XmRString,    "cart6.mdr"   )
#endif
#if IF1_DRIVES > 6
	RES2("if1Drive7",      "If1Drive7",      XmRString,  XPointer,     if1Carts[6],      XmRString,    "cart7.mdr"   )
#endif
#if IF1_DRIVES > 7
	RES2("if1Drive8",      "If1Drive8",      XmRString,  XPointer,     if1Carts[7],      XmRString,    "cart8.mdr"   )
#endif
#endif
#ifdef XZX_MF
	RES2("mf",             "Mf",             XmRBoolean, Boolean,      mf1,              XmRImmediate, FALSE         )
	RES2("mf1Rom",         "Mf1Rom",         XmRString,  XPointer,     mf1Rom,           XmRString,    "mf1.rom"     )
	RES2("mf128Rom",       "Mf128Rom",       XmRString,  XPointer,     mf128Rom,         XmRString,    "mf128.rom"   )
	RES2("mf3Rom",         "Mf3Rom",         XmRString,  XPointer,     mf3Rom,           XmRString,    "mf3.rom"     )
#endif
#ifdef XZX_TRDOS
	RES2("trdos",          "Trdos",          XmRBoolean, Boolean,      trdos,            XmRImmediate, FALSE         )
	RES2("trdosRom",       "TrdosRom",       XmRString,  XPointer,     trdosRom,         XmRString,    "trdos.rom"   )
	RES2("trdosDriveA",    "TrdosDriveA",    XmRString,  XPointer,     trdosDisks[0],    XmRString,    "diska.trd"   )
	RES2("trdosDriveB",    "TrdosDriveB",    XmRString,  XPointer,     trdosDisks[1],    XmRString,    "diskb.trd"   )
	RES2("trdosDriveC",    "TrdosDriveC",    XmRString,  XPointer,     trdosDisks[2],    XmRString,    "diskc.trd"   )
	RES2("trdosDriveD",    "TrdosDriveD",    XmRString,  XPointer,     trdosDisks[3],    XmRString,    "diskd.trd"   )
#endif
#ifdef XZX_MGT
	RES2("mgt",            "Mgt",            XmRBoolean, Boolean,      mgt,              XmRImmediate, FALSE         )
	RES2("mgtRom",         "MgtRom",         XmRString,  XPointer,     mgtRom,           XmRString,    "plusd.rom"   )
	RES2("mgtDriveA",      "MgtDriveA",      XmRString,  XPointer,     mgtDisks[0],      XmRString,    "diska.mgt"   )
	RES2("mgtDriveB",      "MgtDriveB",      XmRString,  XPointer,     mgtDisks[1],      XmRString,    "diskb.mgt"   )
#endif
#ifdef XZX_KMOUSE
	RES2("kmouse",         "Kmouse",         XmRBoolean, Boolean,      kmouse,           XmRImmediate, FALSE         )
	RES2("kmouseGrab",     "KmouseGrab",     XmRBoolean, Boolean,      kmouseGrab,       XmRImmediate, FALSE         )
#endif
};
static KeySym defaultKeys[NUM_KEYMAPPINGS] = {
	XK_q, XK_a, XK_o, XK_p, XK_space
};
static XtResource resourceKeys[] = {
	RES3("joyUp",    "JoyUp",    XmRString, XPointer, 0 * sizeof(char *), XmRString, "q"     )
	RES3("joyDown",  "JoyDown",  XmRString, XPointer, 1 * sizeof(char *), XmRString, "a"     )
	RES3("joyLeft",  "JoyLeft",  XmRString, XPointer, 2 * sizeof(char *), XmRString, "o"     )
	RES3("joyRight", "JoyRight", XmRString, XPointer, 3 * sizeof(char *), XmRString, "p"     )
	RES3("joyFire",  "JoyFire",  XmRString, XPointer, 4 * sizeof(char *), XmRString, "space" )
	RES3("machine",  "Machine",  XmRString, XPointer, 5 * sizeof(char *), XmRString, "48"    )
};
static const char *mach2str[] = {
	"48-2",
	"48",
	"128",
	"3",
#ifdef XZX_PENTAGON
	"Pentagon",
#endif
#ifdef XZX_SCORPION
	"Scorpion",
#endif
#ifdef XZX_DIDAKTIK
	"Didaktik",
#endif
};
void
ResourceRead(int *argcp, char **argvp)
{
	const char *mach;
	char *specialKeys[NELEM(resourceKeys)];
	int i;
	if (*argcp > 1)
	{
		if (argvp[1][0] == '-')
		{
			if (strcmp(argvp[1], "-version") == 0)
			{
				resourcePrintCopyright();
				resourcePrintVersion();
			}
			else
			{
				if (strcmp(argvp[1], "-help") == 0)
				{
					(void)fprintf(stderr, "Usage: %s [options] [snapshot file]\n", ProgName);
				}
				else
				{
					(void)fprintf(stderr, "Unknown option \"%s\".\n", argvp[1]);
				}
				(void)fprintf(stderr, "For available options see manual page.\n");
			}
			Quit(1);
		}
	}
	resourcePrintCopyright();
	XtVaGetApplicationResources(XwinDatas.topLevel, &ResourceDatas, resourceDefaults, XtNumber(resourceDefaults), XtNull);
	XtVaGetApplicationResources(XwinDatas.topLevel, &specialKeys, resourceKeys, XtNumber(resourceKeys), XtNull);
	mach = specialKeys[NELEM(specialKeys) - 1];
	for (i = 0; i < NELEM(mach2str); i++)
	{
		if (strcasecmp(mach, mach2str[i]) == 0)
		{
			goto found_machine;
		}
	}
	Msg(M_WARN, "invalid emulation mode %s, using 48", mach);
	i = SP_48_3;
  found_machine:
	CFGSET(machine, i);
	mach = mach2str[i];
	if (CFGGET(speed) < 10 || CFGGET(speed) > 200)
	{
		Msg(M_WARN, "speed value out of range, setting to default 100%");
		CFGSET(speed, 100);
	}
	if (CFGGET(scale) < 1 || CFGGET(scale) > 3)
	{
		Msg(M_WARN, "scale factor out of range, setting to default 1");
		CFGSET(scale, 1);
	}
	if (CFGGET(refresh) < 1 || CFGGET(refresh) > 50)
	{
		Msg(M_WARN, "refresh rate out of range, setting to default 3");
		CFGSET(refresh, 3);
	}
#ifdef REAL_AUDIO
	if (CFGGET(tapeFastload) != 0 && CFGGET(tapeNoise) != 0)
	{
		Msg(M_WARN, "loading noise in fast mode not supported, disabling");
		CFGSET(tapeNoise, 0);
	}
#endif
#ifdef JOY
	if (CFGGET(joyTolerance) < 1 || CFGGET(joyTolerance) > 500)
	{
		Msg(M_WARN, "joystick tolerance out of range, setting to default 20");
		CFGSET(joyTolerance, 20);
	}
	if (CFGGET(joyAutofire) < 1 || CFGGET(joyAutofire) > 50)
	{
		Msg(M_WARN, "autofire speed out of range, setting to default 25");
		CFGSET(joyAutofire, 25);
	}
	if (CFGGET(joyJuggle) < 1 || CFGGET(joyJuggle) > 50)
	{
		Msg(M_WARN, "juggle speed out of range, setting to default 25");
		CFGSET(joyJuggle, 25);
	}
#endif
	for (i = KM_JOY_UP; i <= KM_JOY_FIRE; i++)
	{
		CFGSET(joyKeys[i], XStringToKeysym(specialKeys[i]));
		if (CFGGET(joyKeys[i]) == NoSymbol)
		{
			Msg(M_WARN, "unknown keysym <%s> found, using <%s>", specialKeys[i], XKeysymToString(defaultKeys[i]));
			CFGSET(joyKeys[i], defaultKeys[i]);
		}
	}
#ifdef REAL_AUDIO
	if (CFGGET(audioSample) < 8000 || CFGGET(audioSample) > 48000)
	{
		Msg(M_WARN, "audio sample rate out of range, setting to default 8000");
		CFGSET(audioSample, 8000);
	}
	if (CFGGET(audioStereo) < AY_MONO || CFGGET(audioStereo) > AY_ABC)
	{
		Msg(M_WARN, "invalid AY mode, using mono");
		CFGSET(audioStereo, AY_MONO);
	}
#endif
#ifdef XBELL_AUDIO
	if (CFGGET(audioVolume) > 100)
	{
		Msg(M_WARN, "audio volume out of range, setting to default 50");
		CFGSET(audioVolume, 50);
	}
#endif
	for (i = 0; i < NELEM(CFGGET(sp3Disks)); i++)
	{
		CFGSET(sp3Disks[i], Strdup(CFGGET(sp3Disks[i]), "ResourceRead"));
	}
#ifdef XZX_IF1
	for (i = 0; i < NELEM(CFGGET(if1Carts)); i++)
	{
		CFGSET(if1Carts[i], Strdup(CFGGET(if1Carts[i]), "ResourceRead"));
	}
	if (CFGGET(if1) > 0 && CFGGET(machine) > SP_128)
	{
		Msg(M_INFO, "Interface I not supported by the %s, disabling", mach);
		CFGSET(if1, 0);
	}
#endif
#ifdef XZX_MF
	if (CFGGET(mf1) > 0 && CFGGET(machine) > SP_3)
	{
		Msg(M_INFO, "Multiface not supported by the %s, disabling", mach);
		CFGSET(mf1, 0);
	}
#endif
#ifdef XZX_TRDOS
	for (i = 0; i < NELEM(CFGGET(trdosDisks)); i++)
	{
		CFGSET(trdosDisks[i], Strdup(CFGGET(trdosDisks[i]), "ResourceRead"));
	}
	if (CFGGET(trdos) > 0 && CFGGET(machine) > SP_128)
	{
		Msg(M_INFO, "Beta 128 not supported by the %s, disabling", mach);
		CFGSET(trdos, 0);
	}
#ifdef XZX_MF
	if (CFGGET(trdos) > 0 && CFGGET(mf1) > 0)
	{
		Msg(M_INFO, "Beta 128 and Multiface not supported, disabling");
		CFGSET(mf1, 0);
	}
#endif
#ifdef XZX_MGT
	if (CFGGET(trdos) > 0 && CFGGET(mgt) > 0)
	{
		Msg(M_INFO, "Beta 128  and +D not supported, disabling");
		CFGSET(mgt, 0);
	}
#endif
#endif
#ifdef XZX_MGT
	for (i = 0; i < NELEM(CFGGET(mgtDisks)); i++)
	{
		CFGSET(mgtDisks[i], Strdup(CFGGET(mgtDisks[i]), "ResourceRead"));
	}
	if (CFGGET(mgt) > 0 && CFGGET(machine) > SP_128)
	{
		Msg(M_INFO, "+D not supported by the %s, disabling", mach);
		CFGSET(mgt, 0);
	}
#ifdef XZX_IF1
	if (CFGGET(mgt) > 0 && CFGGET(if1) > 0)
	{
		Msg(M_INFO, "+D and Interface I not supported, disabling");
		CFGSET(if1, 0);
	}
#endif
#ifdef XZX_MF
	if (CFGGET(mgt) > 0 && CFGGET(mf1) > 0)
	{
		Msg(M_INFO, "+D and Multiface not supported, disabling");
		CFGSET(mf1, 0);
	}
#endif
#endif
#ifdef XZX_KMOUSE
	if (CFGGET(kmouse) > 0  && !PENTAGON_OR_SCORPION)
	{
		Msg(M_INFO, "Kempston Mouse not supported by the %s, disabling", mach);
		CFGSET(kmouse, 0);
	}
#endif
#ifdef XZX_MF
	CFGSET(mf128, CFGGET(mf1));
	CFGSET(mf3, CFGGET(mf1));
#endif
#ifdef DEBUG
	if ((CFGGET(debug) & D_X11) != 0)
	{
		resourcePrint();
	}
#endif
}
static void
resourcePrintCopyright(void)
{
	(void)fprintf(stderr,
	"XZX is copyright (c) 1993,1994 Des Herriott\n"
	"                 (c) 1995-2000 Erik Kunze\n"
	"XZX comes with no warranty; see the file COPYRIGHT for details.\n\n");
}
static void
resourcePrintVersion(void)
{
	(void)fprintf(stderr, "This is %s", Version);
	(void)fprintf(stderr, " with the following compile time options:\n"
	" *  XZX is built for a"
#ifndef WORDS_BIGENDIAN
	" little-endian architecture.\n"
#else
	" big-endian architecture.\n"
#endif
#ifdef JOY
	" *  Analogue joystick emulation is possible.\n"
#endif
#ifdef REAL_AUDIO
	" *  Speaker audio emulation is enabled.\n"
#ifdef REGISTERED
	" *  AY-8912 audio emulation (ACB/ABC stereo) is enabled.\n"
#else
	" *  AY-8912 audio emulation is enabled.\n"
#endif
#endif
#ifdef XBELL_AUDIO
	" *  BEEP emulation via X server bell is enabled.\n"
#endif
#ifdef XZX_PENTAGON
	" *  Pentagon emulation is enabled.\n"
#endif
#ifdef XZX_SCORPION
	" *  Scorpion emulation is enabled.\n"
#endif
#ifdef XZX_DIDAKTIK
	" *  Didaktik emulation is enabled.\n"
#endif
#ifdef XZX_IF1
	" *  Interface 1, microdrive & RS232 emulation are enabled.\n"
#endif
#ifdef XZX_MF
	" *  Multiface emulation is enabled.\n"
#endif
#ifdef XZX_TRDOS
	" *  Beta 128 emulation is enabled.\n"
#endif
#ifdef XZX_MGT
	" *  +D emulation is enabled.\n"
#endif
#ifdef XZX_KMOUSE
	" *  Kempston Mouse emulation is enabled.\n"
#endif
#ifdef PSEUDO_IO
	" *  Character input/output is emulated via ports.\n"
#endif
#ifdef DEBUG
	" *  Debugging is possible.\n"
#endif
	);
}
#ifdef DEBUG
static void
resourcePrint(void)
{
#ifdef XZX_IF1
	int i;
#endif
	Msg(M_DEBUG, "libDir        : %s", CFGGET(libDir));
	Msg(M_DEBUG, "infoDir       : %s", CFGGET(infoDir));
	Msg(M_DEBUG, "pokeDir       : %s", CFGGET(pokeDir));
	Msg(M_DEBUG, "quiet         : %s", BOOL(CFGGET(quiet)));
	Msg(M_DEBUG, "core          : %s", BOOL(CFGGET(core)));
	Msg(M_DEBUG, "monocrome     : %s", BOOL(CFGGET(monocrome)));
	Msg(M_DEBUG, "private       : %s", BOOL(CFGGET(private)));
	Msg(M_DEBUG, "xsync         : %s", BOOL(CFGGET(xsync)));
#ifdef DEBUG
	Msg(M_DEBUG, "debug         : %u", CFGGET(debug));
#endif
	Msg(M_DEBUG, "machine       : %s", mach2str[CFGGET(machine)]);
	Msg(M_DEBUG, "speed         : %d", CFGGET(speed));
	Msg(M_DEBUG, "fastmode      : %s", BOOL(CFGGET(fastmode)));
	Msg(M_DEBUG, "warn48        : %s", BOOL(CFGGET(warn48)));
	Msg(M_DEBUG, "scale         : %d", CFGGET(scale));
	Msg(M_DEBUG, "refresh       : %d", CFGGET(refresh));
	Msg(M_DEBUG, "cache         : %s", BOOL(CFGGET(cache)));
#ifndef NO_ED_TRAPS
	Msg(M_DEBUG, "tapeQuickload : %s", BOOL(CFGGET(tapeQuickload)));
#endif
	Msg(M_DEBUG, "tapeFastload  : %s", BOOL(CFGGET(tapeFastload)));
	Msg(M_DEBUG, "tapeAutostop  : %s", BOOL(CFGGET(tapeAutostop)));
#ifdef REAL_AUDIO
	Msg(M_DEBUG, "tapeNoise     : %s", BOOL(CFGGET(tapeNoise)));
#endif
	Msg(M_DEBUG, "joyUp         : %s", KEY(CFGGET(joyUp)));
	Msg(M_DEBUG, "joyDown       : %s", KEY(CFGGET(joyDown)));
	Msg(M_DEBUG, "joyLeft       : %s", KEY(CFGGET(joyLeft)));
	Msg(M_DEBUG, "joyRight      : %s", KEY(CFGGET(joyRight)));
	Msg(M_DEBUG, "joyFire       : %s", KEY(CFGGET(joyFire)));
#ifdef JOY
	Msg(M_DEBUG, "joy           : %s", BOOL(CFGGET(joy)));
	Msg(M_DEBUG, "joyDev        : %s", CFGGET(joyDev));
	Msg(M_DEBUG, "joyTolerance  : %d", CFGGET(joyTolerance));
	Msg(M_DEBUG, "joyAutofire   : %d", CFGGET(joyAutofire));
	Msg(M_DEBUG, "joyJuggle     : %d", CFGGET(joyJuggle));
#endif
#ifdef AUDIO
	Msg(M_DEBUG, "audio         : %s", BOOL(CFGGET(audio)));
#ifdef REAL_AUDIO
#ifdef NAS_AUDIO
	Msg(M_DEBUG, "audioServer   : %s", CFGGET(audioDev));
#else
	Msg(M_DEBUG, "audioDev      : %s", CFGGET(audioDev));
#endif
	Msg(M_DEBUG, "audioSample   : %d", CFGGET(audioSample));
	Msg(M_DEBUG, "audioStereo   : %d", CFGGET(audioStereo));
#endif
#ifdef XBELL_AUDIO
	Msg(M_DEBUG, "audioVolume   : %u", CFGGET(audioVolume));
#endif
#endif
	Msg(M_DEBUG, "sp48Rom       : %s", CFGGET(sp48Rom));
	Msg(M_DEBUG, "sp128Rom      : %s", CFGGET(sp128Rom));
	Msg(M_DEBUG, "sp3Rom        : %s", CFGGET(sp3Rom));
	Msg(M_DEBUG, "sp3DriveA     : %s", CFGGET(sp3Disks[0]));
	Msg(M_DEBUG, "sp3DriveB     : %s", CFGGET(sp3Disks[1]));
#ifdef XZX_PENTAGON
	Msg(M_DEBUG, "pentagonRom   : %s", CFGGET(pentagonRom));
#endif
#ifdef XZX_SCORPION
	Msg(M_DEBUG, "scorpionRom   : %s", CFGGET(scorpionRom));
#endif
#ifdef XZX_DIDAKTIK
	Msg(M_DEBUG, "didaktikRom   : %s", CFGGET(didaktikRom));
	Msg(M_DEBUG, "didaktikDriveA: %s", CFGGET(didaktikDisks[0]));
	Msg(M_DEBUG, "didaktikDriveB: %s", CFGGET(didaktikDisks[1]));
#endif
#ifdef XZX_IF1
	Msg(M_DEBUG, "if1           : %s", BOOL(CFGGET(if1)));
	Msg(M_DEBUG, "if1Rom        : %s", CFGGET(if1Rom));
	Msg(M_DEBUG, "if1Crlf       : %s", BOOL(CFGGET(if1Crlf)));
	Msg(M_DEBUG, "if1Strcr      : %s", BOOL(CFGGET(if1Strcr)));
	for (i = 0; i < IF1_DRIVES; i++)
	{
		Msg(M_DEBUG, "if1Drive%d     : %s", i,  CFGGET(if1Carts[i]));
	}
#endif
#ifdef XZX_MF
	Msg(M_DEBUG, "mf            : %s", BOOL(CFGGET(mf1)));
	Msg(M_DEBUG, "mf1Rom        : %s", CFGGET(mf1Rom));
	Msg(M_DEBUG, "mf128Rom      : %s", CFGGET(mf128Rom));
	Msg(M_DEBUG, "mf3Rom        : %s", CFGGET(mf3Rom));
#endif
#ifdef XZX_TRDOS
	Msg(M_DEBUG, "trdos         : %s", BOOL(CFGGET(trdos)));
	Msg(M_DEBUG, "trdosRom      : %s", CFGGET(trdosRom));
	Msg(M_DEBUG, "trdosDriveA   : %s", CFGGET(trdosDisks[0]));
	Msg(M_DEBUG, "trdosDriveB   : %s", CFGGET(trdosDisks[1]));
	Msg(M_DEBUG, "trdosDriveC   : %s", CFGGET(trdosDisks[2]));
	Msg(M_DEBUG, "trdosDriveD   : %s", CFGGET(trdosDisks[3]));
#endif
#ifdef XZX_MGT
	Msg(M_DEBUG, "mgt           : %s", BOOL(CFGGET(mgt)));
	Msg(M_DEBUG, "mgtRom        : %s", CFGGET(mgtRom));
	Msg(M_DEBUG, "mgtDriveA     : %s", CFGGET(mgtDisks[0]));
	Msg(M_DEBUG, "mgtDriveB     : %s", CFGGET(mgtDisks[1]));
#endif
#ifdef XZX_KMOUSE
	Msg(M_DEBUG, "kmouse        : %s", BOOL(CFGGET(kmouse)));
	Msg(M_DEBUG, "kmouseGrab    : %s", BOOL(CFGGET(kmouseGrab)));
#endif
}
#endif
int
ResourceWrite(void)
{
	FILE *fp;
	char path[MAXPATHLEN];
	char *home = getenv("HOME");
	int i;
	char *file;
	if (home == (char *)0)
	{
		Msg(M_WARN, "environment variable HOME is not set");
		return -1;
	}
	(void)sprintf(path, "%s/%s", home, RESOURCE_FILE);
	if ((fp = Fopen(path, "wb")) == (FILE *)0)
	{
		Msg(M_PERR, "couldn't open <%s> for writing", path);
		return -1;
	}
	(void)fprintf(fp,
				  "! XZX Preferences File\n"
				  "! Version: " VERSION " (" HOST ")\n"
				  "! This is a generated file!  Do not edit.\n\n");
	(void)fprintf(fp,
				  "Xzx*libDir: %s\n"
				  "Xzx*infoDir: %s\n"
				  "Xzx*pokeDir: %s\n"
				  "Xzx*quiet: %s\n"
				  "Xzx*monocrome: %s\n"
				  "Xzx*private: %s\n",
				  CFGGET(libDir),
				  CFGGET(infoDir),
				  CFGGET(pokeDir),
				  BOOL(CFGGET(quiet)),
				  BOOL(CFGGET(monocrome)),
				  BOOL(CFGGET(private)));
	(void)fprintf(fp,
				  "Xzx*xsync: %s\n",
				  BOOL(CFGGET(xsync)));
#ifdef DEBUG
	(void)fprintf(fp,
				  "Xzx*debug: %u\n",
				  CFGGET(debug));
#endif
	(void)fprintf(fp,
				  "Xzx*machine: %s\n"
				  "Xzx*speed: %d\n"
				  "Xzx*fastmode: %s\n"
				  "Xzx*warn48: %s\n"
				  "Xzx*scale: %d\n"
				  "Xzx*refresh: %d\n"
				  "Xzx*cache: %s\n",
				  mach2str[CFGGET(machine)],
				  CFGGET(speed),
				  BOOL(CFGGET(fastmode)),
				  BOOL(CFGGET(warn48)),
				  CFGGET(scale),
				  CFGGET(refresh),
				  BOOL(CFGGET(cache)));
#ifndef NO_ED_TRAPS
	(void)fprintf(fp,
				  "Xzx*tapeQuickload: %s\n",
				  BOOL(CFGGET(tapeQuickload)));
#endif
	(void)fprintf(fp,
				  "Xzx*tapeFastload: %s\n"
				  "Xzx*tapeAutostop: %s\n",
				  BOOL(CFGGET(tapeFastload)),
				  BOOL(CFGGET(tapeAutostop)));
#ifdef REAL_AUDIO
	(void)fprintf(fp,
				  "Xzx*tapeNoise: %s\n",
				  BOOL(CFGGET(tapeNoise)));
#endif
	(void)fprintf(fp,
				  "Xzx*joyUp: %s\n"
				  "Xzx*joyDown: %s\n"
				  "Xzx*joyLeft: %s\n"
				  "Xzx*joyRight: %s\n"
				  "Xzx*joyFire: %s\n",
				  KEY(CFGGET(joyUp)),
				  KEY(CFGGET(joyDown)),
				  KEY(CFGGET(joyLeft)),
				  KEY(CFGGET(joyRight)),
				  KEY(CFGGET(joyFire)));
#ifdef JOY
	(void)fprintf(fp,
				  "Xzx*joy: %s\n"
				  "Xzx*joyDev: %s\n"
				  "Xzx*joyTolerance: %d\n"
				  "Xzx*joyAutofire: %d\n"
				  "Xzx*joyJuggle: %d\n",
				  BOOL(CFGGET(joy)),
				  CFGGET(joyDev),
				  CFGGET(joyTolerance) > 0 ? CFGGET(joyTolerance) : 20,
				  CFGGET(joyAutofire) > 0 ? CFGGET(joyAutofire) : 25,
				  CFGGET(joyJuggle) > 0 ? CFGGET(joyJuggle) : 25);
#endif
#ifdef AUDIO
	(void)fprintf(fp,
				  "Xzx*audio: %s\n",
				  BOOL(CFGGET(audio)));
#ifdef REAL_AUDIO
#ifdef NAS_AUDIO
	(void)fprintf(fp,
				  "Xzx*audioServer: %s\n",
				  CFGGET(audioDev));
#else
	(void)fprintf(fp,
				  "Xzx*audioDev: %s\n",
				  CFGGET(audioDev));
#endif
	(void)fprintf(fp,
				  "Xzx*audioSample: %d\n"
				  "Xzx*audioStereo: %d\n",
				  CFGGET(audioSample),
				  CFGGET(audioStereo));
#endif
#ifdef XBELL_AUDIO
	(void)fprintf(fp,
				  "Xzx*audioVolume: %u\n",
				  CFGGET(audioVolume));
#endif
#endif
	(void)fprintf(fp,
				  "Xzx*sp48Rom: %s\n"
				  "Xzx*sp128Rom: %s\n"
				  "Xzx*sp3Rom: %s\n",
				  CFGGET(sp48Rom),
				  CFGGET(sp128Rom),
				  CFGGET(sp3Rom));
	for (i = 0; i < NELEM(CFGGET(sp3Disks)); i++)
	{
		if ((file = FdcDiskName(i)) != (char *)0)
		{
			(void)fprintf(fp,
						  "Xzx*sp3Drive%c: %s\n",
						  (char)('A' + i), file);
		}
	}
#ifdef XZX_PENTAGON
	(void)fprintf(fp,
				  "Xzx*pentagonRom: %s\n",
				  CFGGET(pentagonRom));
#endif
#ifdef XZX_SCORPION
	(void)fprintf(fp,
				  "Xzx*scorpionRom: %s\n",
				  CFGGET(scorpionRom));
#endif
#ifdef XZX_DIDAKTIK
	(void)fprintf(fp,
				  "Xzx*didaktikRom: %s\n",
				  CFGGET(didaktikRom));
	for (i = 0; i < NELEM(CFGGET(didaktikDisks)); i++)
	{
		if ((file = DidaktikDiskName(i)) != (char *)0)
		{
			(void)fprintf(fp,
						  "Xzx*didaktikDrive%c: %s",
						  (char)('A' + i), file);
		}
	}
#endif
#ifdef XZX_IF1
	(void)fprintf(fp,
				  "Xzx*if1: %s\n"
				  "Xzx*if1Rom: %s\n"
				  "Xzx*if1Crlf: %s\n"
				  "Xzx*if1Strcr: %s\n",
				  BOOL(CFGGET(if1)),
				  CFGGET(if1Rom),
				  BOOL(CFGGET(if1Crlf)),
				  BOOL(CFGGET(if1Strcr)));
	for (i = 0; i < IF1_DRIVES; i++)
	{
		if ((file = If1CartName(i)) != (char *)0)
		{
			(void)fprintf(fp,
						  "Xzx*if1Drive%d: %s\n",
						  i, file);
		}
	}
#endif
#ifdef XZX_MF
	(void)fprintf(fp,
				  "Xzx*mf: %s\n"
				  "Xzx*mf1Rom: %s\n"
				  "Xzx*mf128Rom: %s\n"
				  "Xzx*mf3Rom: %s\n",
				  BOOL((CFGGET(mf1) | CFGGET(mf128) | CFGGET(mf3))),
				  CFGGET(mf1Rom),
				  CFGGET(mf128Rom),
				  CFGGET(mf3Rom));
#endif
#ifdef XZX_TRDOS
	(void)fprintf(fp,
				  "Xzx*trdos: %s\n"
				  "Xzx*trdosRom: %s\n",
				  BOOL(CFGGET(trdos)),
				  CFGGET(trdosRom));
	for (i = 0; i < NELEM(CFGGET(trdosDisks)); i++)
	{
		if ((file = TrdosDiskName(i)) != (char *)0)
		{
			(void)fprintf(fp,
						  "Xzx*trdosDrive%c: %s\n",
						  (char)('A' + i), file);
		}
	}
#endif
#ifdef XZX_MGT
	(void)fprintf(fp,
				  "Xzx*mgt: %s\n"
				  "Xzx*mgtRom: %s\n",
				  BOOL(CFGGET(mgt)),
				  CFGGET(mgtRom));
	for (i = 0; i < NELEM(CFGGET(mgtDisks)); i++)
	{
		if ((file = MgtDiskName(i)) != (char *)0)
		{
			(void)fprintf(fp,
						  "Xzx*mgtDrive%c: %s\n",
						  (char)('A' + i), file);
		}
	}
#endif
#ifdef XZX_KMOUSE
	(void)fprintf(fp,
				  "Xzx*kmouse: %s\n"
				  "Xzx*kmouseGrab: %s\n",
				  BOOL(CFGGET(kmouse)),
				  BOOL(CFGGET(kmouseGrab)));
#endif
	(void)fprintf(fp, "\n! EOF\n");
	(void)fclose(fp);
	return 0;
}
