/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
static char rcsid[] = "$Id: sbar.c,v 1.11 2000/10/29 19:06:57 erik Rel $";
#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/LabelG.h>
#include <stdio.h>
#include <assert.h>
#include "sbar.h"
typedef struct _sbarData {
	Widget model;
	Widget speed;
	Widget led;
	Pixel onColor, offColor;
	XtIntervalId timer;
} sbarData;
#define SBAR_LED_DELAY					100
#define XtNull 							(XtPointer)0
static void sbarOffLed(XtPointer, XtIntervalId);
static sbarData sbarDatas;
Widget
SbarCreateBar(Widget parent)
{
	sbarData *sdp = &sbarDatas;
	Widget bar, frame;
	Arg arg[10];
	Cardinal i;
	i = 0;
	XtSetArg(arg[i], XmNmarginHeight, 2  ); i++;
	XtSetArg(arg[i], XmNmarginWidth,  2  ); i++;
	XtSetArg(arg[i], XmNfractionBase, 100); i++;
	assert(i <= XtNumber(arg));
	bar = XmCreateForm(parent, (char *)"sbar", arg, i);
	XtManageChild(bar);
	i = 0;
	XtSetArg(arg[i], XmNalignment,        XmALIGNMENT_BEGINNING); i++;
	XtSetArg(arg[i], XmNmarginWidth,      0                    ); i++;
	XtSetArg(arg[i], XmNleftAttachment,   XmATTACH_FORM        ); i++;
	XtSetArg(arg[i], XmNleftOffset,       5                    ); i++;
	XtSetArg(arg[i], XmNtopAttachment,    XmATTACH_FORM        ); i++;
	XtSetArg(arg[i], XmNbottomAttachment, XmATTACH_FORM        ); i++;
	assert(i <= XtNumber(arg));
	sdp->model = XmCreateLabelGadget(bar, (char *)"model", arg, i);
	XtManageChild(sdp->model);
	i = 0;
	XtSetArg(arg[i], XmNalignment,        XmALIGNMENT_BEGINNING); i++;
	XtSetArg(arg[i], XmNmarginWidth,      0                    ); i++;
	XtSetArg(arg[i], XmNmarginRight,      5                    ); i++;
	XtSetArg(arg[i], XmNleftAttachment,   XmATTACH_WIDGET      ); i++;
	XtSetArg(arg[i], XmNleftWidget,       sdp->model           ); i++;
	XtSetArg(arg[i], XmNtopAttachment,    XmATTACH_FORM        ); i++;
	XtSetArg(arg[i], XmNbottomAttachment, XmATTACH_FORM        ); i++;
	assert(i <= XtNumber(arg));
	sdp->speed = XmCreateLabelGadget(bar, (char *)"speed", arg, i);
	XtManageChild(sdp->speed);
	i = 0;
	XtSetArg(arg[i], XmNmappedWhenManaged, False             ); i++;
	XtSetArg(arg[i], XmNshadowType,        XmSHADOW_IN       ); i++;
	XtSetArg(arg[i], XmNmarginHeight,      0                 ); i++;
	XtSetArg(arg[i], XmNmarginWidth,       0                 ); i++;
	XtSetArg(arg[i], XmNrightAttachment,   XmATTACH_FORM     ); i++;
	XtSetArg(arg[i], XmNrightOffset,       5                 ); i++;
	XtSetArg(arg[i], XmNtopAttachment,     XmATTACH_POSITION ); i++;
	XtSetArg(arg[i], XmNtopPosition,       20                ); i++;
	XtSetArg(arg[i], XmNbottomAttachment,  XmATTACH_POSITION ); i++;
	XtSetArg(arg[i], XmNbottomPosition,    80                ); i++;
	assert(i <= XtNumber(arg));
	frame = XmCreateFrame(bar, (char *)"frame", arg, i);
	XtManageChild(frame);
	i = 0;
	XtSetArg(arg[i], XmNlabelType, XmPIXMAP); i++;
	XtSetArg(arg[i], XmNwidth,     15      ); i++;
	assert(i <= XtNumber(arg));
	sdp->led = XmCreateLabelGadget(frame, (char *)"led", arg, i);
	XtManageChild(sdp->led);
	XtVaGetValues(sdp->led,
				  XmNforeground, &(sdp->onColor),
				  XmNbackground, &(sdp->offColor),
				  XtNull);
	return bar;
}
void
SbarSetModel(const char *name)
{
	XmString str = XmStringCreateLtoR((char *)name, XmFONTLIST_DEFAULT_TAG);
	XtVaSetValues(sbarDatas.model, XmNlabelString, str, XtNull);
	XmStringFree(str);
}
void
SbarSetSpeed(int speed, int fps)
{
	static int prevSpeed = -1, prevFps = -1;
	char buf[80];
	XmString str;
	if (speed != prevSpeed || fps != prevFps)
	{
		prevSpeed = speed;
		prevFps = fps;
		(void)sprintf(buf, " at %d%%, %d fps", speed, fps);
		str = XmStringCreateLtoR(buf, XmFONTLIST_DEFAULT_TAG);
		XtVaSetValues(sbarDatas.speed, XmNlabelString, str, XtNull);
		XmStringFree(str);
	}
}
void
SbarSetLed(enum LedState state)
{
	sbarData *sdp = &sbarDatas;
	switch (state)
	{
		case LED_SHOW:
			XtMapWidget(XtParent(sdp->led));
			break;
		case LED_HIDE:
			XtUnmapWidget(XtParent(sdp->led));
		case LED_OFF:
			if (sdp->timer != (XtIntervalId)0)
			{
				XtRemoveTimeOut(sdp->timer);
				sdp->timer = 0;
			}
			XtVaSetValues(sdp->led, XmNbackground, sdp->offColor, XtNull);
			break;
		case LED_ON:
			if (sdp->timer != (XtIntervalId)0)
			{
				XtRemoveTimeOut(sdp->timer);
			}
			sdp->timer = XtAppAddTimeOut(XtWidgetToApplicationContext(sdp->led), SBAR_LED_DELAY,
										 (XtTimerCallbackProc)sbarOffLed, XtNull);
			XtVaSetValues(sdp->led, XmNbackground, sdp->onColor, XtNull);
			break;
	}
}
static void
sbarOffLed(XtPointer client, XtIntervalId id)
{
	sbarData *sdp = &sbarDatas;
	sdp->timer = 0;
	XtVaSetValues(sdp->led, XmNbackground, sdp->offColor, XtNull);
}
