/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: vtp.c,v 4.38 2000/11/04 18:22:04 erik Rel $";
#endif
#include <Xm/Xm.h>
#include <Xm/CascadeBG.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/List.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/TextF.h>
#include <Xm/ToggleBG.h>
#include <stdlib.h>
#include <assert.h>
#include "util.h"
#include "ui.h"
#include "tape.h"
#include "toolhelp.h"
#include "xwin.h"
#include "vtp.h"
enum { LOAD_ATTACH, LOAD_DETACH, SAVE_ATTACH, SAVE_DETACH, VTP_MAX_MENUS };
typedef struct _vtpMenu {
	const char *label;
	WidgetClass *class;
	int cbvalue;
	Boolean sensitive;
	Widget widget;
} vtpMenu;
enum { PLAYER_REW, PLAYER_STOP, PLAYER_PLAY, PLAYER_FF, VTP_MAX_BUTTONS };
typedef struct _vtpButton {
	const char *label;
	int cbvalue;
	Widget widget;
} vtpButton;
typedef struct _vtpData {
	Widget shell;
	vtpMenu menus[VTP_MAX_MENUS];
	struct {
		Widget name;
		Widget list;
		vtpButton buttons[VTP_MAX_BUTTONS];
		XtIntervalId timer;
	} load;
	struct {
		Widget name;
		Widget list;
	} save;
} vtpData;
enum { W_NONE,
	   W_LOPEN, W_LCLOSE, W_SOPEN, W_SCLOSE, W_QUIT,
	   W_LSELECT,
	   W_FF, W_REW, W_PLAY, W_STOP };
#define XtNull 							(XtPointer)0
#define XtAddCallback(w,cb,f,val)		XtAddCallback(w,cb,f,(XtPointer)(val))
static void vtpCallback(Widget, XtPointer, XtPointer);
static void vtpEventhandler(Widget, XtPointer, XEvent *, Boolean *);
static void vtpChangeBlock(XtPointer, XtIntervalId);
static Widget vtpRetrieveMenuWidget(void);
static void vtpUpdatePosition(void);
static void vtpUpdateButtons(void);
static vtpData vtpDatas = {
	(Widget)0,
	{
		{ "Attach input tape...",  &xmPushButtonGadgetClass, W_LOPEN,  True,  (Widget)0 },
		{ "Detach input tape",     &xmPushButtonGadgetClass, W_LCLOSE, False, (Widget)0 },
		{ "Attach output tape...", &xmPushButtonGadgetClass, W_SOPEN,  True,  (Widget)0 },
		{ "Detach output tape",    &xmPushButtonGadgetClass, W_SCLOSE, False, (Widget)0 },
	},
	{
		(Widget)0,
		(Widget)0,
		{
			{ "REW",  W_REW,  (Widget)0 },
			{ "STOP", W_STOP, (Widget)0 },
			{ "PLAY", W_PLAY, (Widget)0 },
			{ "FF",   W_FF,   (Widget)0 }
		},
		(XtIntervalId)0,
	},
	{
		(Widget)0,
		(Widget)0
	}
};
void
VtpShow(void)
{
	vtpData *vdp = &vtpDatas;
	if (vdp->shell == (Widget)0)
	{
		Widget bar, menu, cascade, frame, frame2, form, form2, w;
		XmString title;
		Arg arg[13];
		Cardinal i;
		int j;
		title = XmStringCreateLtoR((char *)"XZX Tape Manager", XmFONTLIST_DEFAULT_TAG);
		i = 0;
		XtSetArg(arg[i], XmNfractionBase,   100    ); i++;
		XtSetArg(arg[i], XmNnoResize,       True   ); i++;
		XtSetArg(arg[i], XmNautoUnmanage,   False  ); i++;
		XtSetArg(arg[i], XmNdeleteResponse, XmUNMAP); i++;
		XtSetArg(arg[i], XmNdialogTitle,    title  ); i++;
		assert(i <= XtNumber(arg));
		vdp->shell = XmCreateFormDialog(XwinDatas.topLevel, (char *)"fd", arg, i);
		XtAddCallback(vdp->shell, XmNunmapCallback, vtpCallback, W_QUIT);
		XmStringFree(title);
		i = 0;
		XtSetArg(arg[i], XmNleftAttachment,  XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNrightAttachment, XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNtopAttachment,   XmATTACH_FORM); i++;
		assert(i <= XtNumber(arg));
		bar = XmCreateMenuBar(vdp->shell, (char *)"bar", arg, i);
		menu = XmCreatePulldownMenu(bar, (char *)"pd", (ArgList)0, 0);
		i = 0;
		XtSetArg(arg[i], XmNsubMenuId, menu); i++;
		assert(i <= XtNumber(arg));
		cascade = XmCreateCascadeButtonGadget(bar, (char *)"File", arg, i);
		for (j = LOAD_ATTACH; j < VTP_MAX_MENUS; j++)
		{
			w = XtVaCreateManagedWidget(vdp->menus[j].label, *(vdp->menus[j].class), menu,
										XmNsensitive, vdp->menus[j].sensitive, XtNull);
			if (vdp->menus[j].cbvalue != W_NONE)
			{
				XtAddCallback(w, XmNactivateCallback, vtpCallback, vdp->menus[j].cbvalue);
			}
			XtManageChild(w);
			vdp->menus[j].widget = w;
		}
		XtManageChild(cascade);
		XtManageChild(bar);
		frame = UiCreateFrame(vdp->shell, "LOAD");
		XtVaSetValues(frame,
					  XmNleftAttachment, XmATTACH_FORM,
					  XmNleftOffset, 10,
					  XmNrightAttachment, XmATTACH_POSITION,
					  XmNrightPosition, 50,
					  XmNrightOffset, 2,
					  XmNtopAttachment, XmATTACH_WIDGET,
					  XmNtopWidget, bar,
					  XmNtopOffset, 10,
					  XmNbottomAttachment, XmATTACH_FORM,
					  XmNbottomOffset, 10,
					  XtNull);
		i = 0;
		XtSetArg(arg[i], XmNfractionBase, VTP_MAX_BUTTONS); i++;
		assert(i <= XtNumber(arg));
		form = XmCreateForm(frame, (char *)"form", arg, i);
		XtManageChild(form);
		i = 0;
		XtSetArg(arg[i], XmNleftAttachment,  XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNrightAttachment, XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNtopAttachment,   XmATTACH_FORM); i++;
		assert(i <= XtNumber(arg));
		form2 = XmCreateForm(form, (char *)"form", arg, i);
		i = 0;
		XtSetArg(arg[i], XmNleftAttachment,   XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNtopAttachment,    XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNbottomAttachment, XmATTACH_FORM); i++;
		assert(i <= XtNumber(arg));
		w = XmCreateLabel(form2, (char *)"File", arg, i);
		XtManageChild(w);
		i = 0;
		XtSetArg(arg[i], XmNmaxLength,             255            ); i++;
		XtSetArg(arg[i], XmNeditable,              False          ); i++;
		XtSetArg(arg[i], XmNcursorPositionVisible, False          ); i++;
		XtSetArg(arg[i], XmNtraversalOn,           False          ); i++;
		XtSetArg(arg[i], XmNhighlightThickness,    0              ); i++;
		XtSetArg(arg[i], XmNleftAttachment,        XmATTACH_WIDGET); i++;
		XtSetArg(arg[i], XmNleftWidget,            w              ); i++;
		XtSetArg(arg[i], XmNleftOffset,            3              ); i++;
		XtSetArg(arg[i], XmNrightAttachment,       XmATTACH_FORM  ); i++;
		XtSetArg(arg[i], XmNtopAttachment,         XmATTACH_FORM  ); i++;
		XtSetArg(arg[i], XmNbottomAttachment,      XmATTACH_FORM  ); i++;
		assert(i <= XtNumber(arg));
		w = XmCreateTextField(form2, (char *)"text", arg, i);
		XtManageChild(w);
		vdp->load.name = w;
		XtManageChild(form2);
		for (j = PLAYER_REW; j < VTP_MAX_BUTTONS; j++)
		{
			i = 0;
			XtSetArg(arg[i], XmNsensitive,        False            ); i++;
			XtSetArg(arg[i], XmNleftAttachment,   XmATTACH_POSITION); i++;
			XtSetArg(arg[i], XmNleftPosition,     j                ); i++;
			XtSetArg(arg[i], XmNrightAttachment,  XmATTACH_POSITION); i++;
			XtSetArg(arg[i], XmNrightPosition,    j + 1            ); i++;
			XtSetArg(arg[i], XmNbottomAttachment, XmATTACH_FORM    ); i++;
			assert(i <= XtNumber(arg));
			w = XmCreatePushButtonGadget(form, (char *)(vdp->load.buttons[j].label), arg, i);
			if (j == PLAYER_REW || j == PLAYER_FF)
			{
				XtAddCallback(w, XmNarmCallback, vtpCallback, vdp->load.buttons[j].cbvalue);
				XtAddCallback(w, XmNdisarmCallback, vtpCallback, vdp->load.buttons[j].cbvalue);
			}
			else
			{
				XtAddCallback(w, XmNactivateCallback, vtpCallback, vdp->load.buttons[j].cbvalue);
			}
			XtManageChild(w);
			vdp->load.buttons[j].widget = w;
		}
		i = 0;
		XtSetArg(arg[i], XmNlistSizePolicy,         XmVARIABLE                          ); i++;
		XtSetArg(arg[i], XmNselectionPolicy,        XmSINGLE_SELECT                     ); i++;
		XtSetArg(arg[i], XmNscrollBarDisplayPolicy, XmSTATIC                            ); i++;
		XtSetArg(arg[i], XmNscrollHorizontal,       False                               ); i++;
		XtSetArg(arg[i], XmNscrollVertical,         True                                ); i++;
		XtSetArg(arg[i], XmNleftAttachment,         XmATTACH_FORM                       ); i++;
		XtSetArg(arg[i], XmNrightAttachment,        XmATTACH_FORM                       ); i++;
		XtSetArg(arg[i], XmNtopAttachment,          XmATTACH_WIDGET                     ); i++;
		XtSetArg(arg[i], XmNtopWidget,              form2                               ); i++;
		XtSetArg(arg[i], XmNtopOffset,              3                                   ); i++;
		XtSetArg(arg[i], XmNbottomAttachment,       XmATTACH_WIDGET                     ); i++;
		XtSetArg(arg[i], XmNbottomWidget,           vdp->load.buttons[PLAYER_REW].widget); i++;
		XtSetArg(arg[i], XmNbottomOffset,           3                                   ); i++;
		assert(i <= XtNumber(arg));
		w = XmCreateScrolledList(form, (char *)"list", arg, i);
		XtAddCallback(w, XmNsingleSelectionCallback, vtpCallback, W_LSELECT);
		XtAddEventHandler(w, ButtonPressMask | ButtonReleaseMask, False, vtpEventhandler, XtNull);
		XtSetSensitive(XtParent(w), False);
		XtManageChild(w);
		vdp->load.list = w;
		ToolhelpAddWidget(w, "Press left button for selection,\nright button for information.");
		UiFixiateListWidth(vdp->shell, w, 29);
		frame2 = UiCreateFrame(vdp->shell, "SAVE");
		XtVaSetValues(frame2,
					  XmNleftAttachment, XmATTACH_WIDGET,
					  XmNleftWidget, frame,
					  XmNleftOffset, 2,
					  XmNrightAttachment, XmATTACH_FORM,
					  XmNrightOffset, 10,
					  XmNtopAttachment, XmATTACH_WIDGET,
					  XmNtopWidget, bar,
					  XmNtopOffset, 10,
					  XmNbottomAttachment, XmATTACH_FORM,
					  XmNbottomOffset, 10,
					  XtNull);
		form = XmCreateForm(frame2, (char *)"form", (ArgList)0, 0);
		XtManageChild(form);
		i = 0;
		XtSetArg(arg[i], XmNleftAttachment,  XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNrightAttachment, XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNtopAttachment,   XmATTACH_FORM); i++;
		assert(i <= XtNumber(arg));
		form2 = XmCreateForm(form, (char *)"form", arg, i);
		i = 0;
		XtSetArg(arg[i], XmNleftAttachment,   XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNtopAttachment,    XmATTACH_FORM); i++;
		XtSetArg(arg[i], XmNbottomAttachment, XmATTACH_FORM); i++;
		assert(i <= XtNumber(arg));
		w = XmCreateLabel(form2, (char *)"File", arg, i);
		XtManageChild(w);
		i = 0;
		XtSetArg(arg[i], XmNmaxLength,             255            ); i++;
		XtSetArg(arg[i], XmNeditable,              False          ); i++;
		XtSetArg(arg[i], XmNcursorPositionVisible, False          ); i++;
		XtSetArg(arg[i], XmNtraversalOn,           False          ); i++;
		XtSetArg(arg[i], XmNhighlightThickness,    0              ); i++;
		XtSetArg(arg[i], XmNleftAttachment,        XmATTACH_WIDGET); i++;
		XtSetArg(arg[i], XmNleftWidget,            w              ); i++;
		XtSetArg(arg[i], XmNleftOffset,            3              ); i++;
		XtSetArg(arg[i], XmNrightAttachment,       XmATTACH_FORM  ); i++;
		XtSetArg(arg[i], XmNtopAttachment,         XmATTACH_FORM  ); i++;
		XtSetArg(arg[i], XmNbottomAttachment,      XmATTACH_FORM  ); i++;
		assert(i <= XtNumber(arg));
		w = XmCreateTextField(form2, (char *)"text", arg, i);
		XtManageChild(w);
		vdp->save.name = w;
		XtManageChild(form2);
		i = 0;
		XtSetArg(arg[i], XmNvisibleItemCount,       8              ); i++;
		XtSetArg(arg[i], XmNlistSizePolicy,         XmVARIABLE     ); i++;
		XtSetArg(arg[i], XmNselectionPolicy,        XmSINGLE_SELECT); i++;
		XtSetArg(arg[i], XmNscrollBarDisplayPolicy, XmSTATIC       ); i++;
		XtSetArg(arg[i], XmNscrollHorizontal,       False          ); i++;
		XtSetArg(arg[i], XmNscrollVertical,         True           ); i++;
		XtSetArg(arg[i], XmNleftAttachment,         XmATTACH_FORM  ); i++;
		XtSetArg(arg[i], XmNrightAttachment,        XmATTACH_FORM  ); i++;
		XtSetArg(arg[i], XmNtopAttachment,          XmATTACH_WIDGET); i++;
		XtSetArg(arg[i], XmNtopWidget,              form2          ); i++;
		XtSetArg(arg[i], XmNtopOffset,              3              ); i++;
		XtSetArg(arg[i], XmNbottomAttachment,       XmATTACH_FORM  ); i++;
		assert(i <= XtNumber(arg));
		w = XmCreateScrolledList(form, (char *)"list", arg, i);
		XtSetSensitive(XtParent(w), False);
		XtManageChild(w);
		vdp->save.list = w;
		UiFixiateListWidth(vdp->shell, w, 29);
	}
	XtManageChild(vdp->shell);
	XtPopup(XtParent(vdp->shell), XtGrabNone);
}
void
VtpHide(void)
{
	XtUnmanageChild(vtpDatas.shell);
	XtPopdown(XtParent(vtpDatas.shell));
}
static void
vtpCallback(Widget w, XtPointer client, XtPointer call)
{
	vtpData *vdp = &vtpDatas;
	switch ((int)client)
	{
		case W_LOPEN:
			VtpNewInputFile();
			break;
		case W_LCLOSE:
			TapeInDetach();
			XmTextFieldSetString(vdp->load.name, (char *)"");
			XmListDeleteAllItems(vdp->load.list);
			vtpUpdateButtons();
			break;
		case W_SOPEN:
			VtpNewOutputFile();
			break;
		case W_SCLOSE:
			TapeOutDetach();
			XmTextFieldSetString(vdp->save.name, (char *)"");
			XmListDeleteAllItems(vdp->save.list);
			XtSetSensitive(w, False);
			break;
		case W_QUIT:
			XmToggleButtonGadgetSetState(vtpRetrieveMenuWidget(), False, False);
			break;
		case W_LSELECT:
			if ((TapeGetState() & (TS_PLAY | TS_ERRIN)) != 0)
			{
				XmListSelectPos(vdp->load.list, (int)(TapeInGetBlock() + 1), False);
				(void)XBell(XtDisplayOfObject(vdp->shell), 5);
			}
			else
			{
				TapeInSetBlock((unsigned int)((XmListCallbackStruct*)call)->item_position - 1);
				VtpUpdate();
			}
			break;
		case W_REW:
		case W_FF:
			switch (((XmPushButtonCallbackStruct *)call)->reason)
			{
				case XmCR_ARM:
					vtpChangeBlock(client, 1);
					break;
				case XmCR_DISARM:
					if (vdp->load.timer != (XtIntervalId)0)
					{
						XtRemoveTimeOut(vdp->load.timer);
						vdp->load.timer = (XtIntervalId)0;
					}
					vtpUpdateButtons();
					break;
			}
			break;
		case W_PLAY:
			TapeInPlay();
			break;
		case W_STOP:
			TapeInStop();
			break;
	}
}
static void
vtpEventhandler(Widget w, XtPointer client, XEvent *xev, Boolean *cont)
{
	static Widget label = (Widget)0;
	if (xev->xbutton.button == Button3)
	{
		if (xev->type == ButtonPress)
		{
			Widget shell;
			XmString title;
			char *s;
			int pos, item;
			if (label == (Widget)0)
			{
				Arg arg[1];
				Cardinal i;
				shell = XtVaCreatePopupShell("shell", overrideShellWidgetClass, XwinDatas.topLevel,
											 XmNallowShellResize, True,
											 XtNull);
				i = 0;
				XtSetArg(arg[i], XmNalignment, XmALIGNMENT_BEGINNING); i++;
				assert(i <= XtNumber(arg));
				label = XmCreateLabel(shell, (char *)"vtpblockinfo", arg, i);
				XtManageChild(label);
			}
			else
			{
				shell = XtParent(label);
			}
			pos = XmListYToPos(w, xev->xbutton.y);
			XtVaGetValues(w, XmNitemCount, &item, XtNull);
			if (pos > item)
			{
				pos = item;
			}
			s = (char *)TapeInGetInfo((unsigned int)(pos - 1));
			title = XmStringCreateLtoR(s, XmFONTLIST_DEFAULT_TAG);
			free(s);
			XtVaSetValues(label, XmNlabelString, title, XtNull);
			XmStringFree(title);
			XtVaSetValues(shell,
						  XmNx, xev->xbutton.x_root,
						  XmNy, xev->xbutton.y_root,
						  XtNull);
			XtPopup(shell, XtGrabNone);
		}
		else
		{
			XtPopdown(XtParent(label));
		}
	}
}
static void
vtpChangeBlock(XtPointer client, XtIntervalId id)
{
	static unsigned long timerValueStart = 500;
	static unsigned long timerValueNew = 50;
	static unsigned long timerValue;
	static unsigned long timerCounter;
	unsigned int block;
	if (id == 1)
	{
		timerValue = timerValueStart;
		timerCounter = 0;
	}
	else
	{
		timerCounter++;
	}
	if (timerCounter > (5000 / timerValueStart) && timerValue == timerValueStart)
	{
		timerValue = timerValueNew;
		timerCounter = 0;
	}
	if (timerCounter > 5000 / timerValueNew && timerValue == timerValueNew)
	{
		timerCounter = 0;
	}
	block = TapeInGetBlock();
	if (((int)client == W_FF && TapeInSetBlock(block + 1) < 0) || ((int)client == W_REW && TapeInSetBlock(block - 1) < 0))
	{
		(void)XBell(XtDisplayOfObject(vtpDatas.shell), 5);
	}
	else
	{
		vtpUpdatePosition();
		vtpDatas.load.timer= XtAppAddTimeOut(XtWidgetToApplicationContext(vtpDatas.shell), id == 1 ? 500UL : timerValue,
											 (XtTimerCallbackProc)vtpChangeBlock, client);
	}
}
static Widget
vtpRetrieveMenuWidget(void)
{
	static Widget w = (Widget)0;
	if (w == (Widget)0)
	{
		w = XtNameToWidget(XwinDatas.topLevel, "*Tape manager");
		assert(w != (Widget)0);
	}
	return w;
}
void
VtpNewInputFile(void)
{
	int types[4] = { FT_TAP, FT_TZX, FT_VOC, FT_END };
	char *fname;
	XmString *listItems;
	TapeRecord *tr;
	unsigned int blocks;
	vtpData *vdp = &vtpDatas;
	int i;
	if ((fname = UiFileSelector("Input Tape", 1, types)) == (char *)0)
	{
		return;
	}
	if (TapeInAttach(fname) < 0)
	{
		XmListDeleteAllItems(vdp->load.list);
		free(fname);
		fname = (char *)"";
		goto quit;
	}
	if (vdp->shell == (Widget)0 || XtIsManaged(XtParent(vdp->shell)) == False)
	{
		XmToggleButtonGadgetSetState(vtpRetrieveMenuWidget(), True, True);
	}
	blocks = TapeInGetBlocks();
	listItems = (XmString *)XtMalloc(blocks * sizeof(XmString *));
	for (i = 0; i < (int)blocks; i++)
	{
		tr = TapeInGetRecord((unsigned int)i);
		assert(tr !=  (TapeRecord *)0);
		listItems[i] = XmStringCreateLtoR(tr->str, XmFONTLIST_DEFAULT_TAG);
	}
	XtVaSetValues(vdp->load.list,
				  XmNitems, listItems,
				  XmNitemCount, i,
				  XtNull);
	while (--i >= 0)
	{
		XmStringFree(listItems[i]);
	}
	XtFree((char *)listItems);
  quit:
	XmTextFieldSetString(vdp->load.name, (char *)GetBaseName(fname));
	VtpUpdate();
}
void
VtpNewOutputFile(void)
{
	int types[4] = { FT_TAP, FT_END };
	char *fname;
	Boolean menu = True, list = True;
	vtpData *vdp = &vtpDatas;
	if ((fname = UiFileSelector("Output Tape", 0, types)) == (char *)0)
	{
		return;
	}
	if (TapeOutAttach(fname) < 0)
	{
		free(fname);
		fname = (char *)"";
		menu = list = False;
		goto quit;
	}
	if (vdp->shell == (Widget)0 || XtIsManaged(XtParent(vdp->shell)) == False)
	{
		XmToggleButtonGadgetSetState(vtpRetrieveMenuWidget(), True, True);
	}
  quit:
	XtSetSensitive(vdp->menus[SAVE_DETACH].widget, menu);
	XmTextFieldSetString(vdp->save.name, (char *)GetBaseName(fname));
	XmListDeleteAllItems(vdp->save.list);
	XtSetSensitive(vdp->save.list, list);
}
void
VtpUpdate(void)
{
	vtpUpdatePosition();
	vtpUpdateButtons();
}
static void
vtpUpdatePosition(void)
{
	Widget list = vtpDatas.load.list;
	int top, visible;
	int pos = (int)(TapeInGetBlock() + 1);
	XmListSelectPos(list, pos, False);
	XtVaGetValues(list,
				  XmNtopItemPosition, &top,
				  XmNvisibleItemCount, &visible,
				  XtNull);
	if (pos < top)
	{
		XmListSetPos(list, pos);
	}
	else if (pos >= (top + visible))
	{
		XmListSetBottomPos(list, pos);
	}
}
static void
vtpUpdateButtons(void)
{
	Boolean list = True, play = False, stop = False, ff = False, rew = False, open = False, close = False;
	unsigned int state = TapeGetState();
	unsigned int block = TapeInGetBlock();
	vtpData *vdp = &vtpDatas;
	if ((state & TS_PLAY) != 0)
	{
		stop = True;
	}
	else if ((state & (TS_TAPIN | TS_TZXIN | TS_VOCIN)) != 0)
	{
		play = (state & (TS_ERRIN | TS_ENDIN)) == 0 ? True : False;
		if ((state & TS_ERRIN) == 0)
		{
			ff = (block + 1) < TapeInGetBlocks() ? True : False;
			rew= block > 0 ? True : False;
		}
		open = True;
		close = True;
	}
	else
	{
		list = False;
		open = True;
	}
	XtSetSensitive(XtParent(vdp->load.list), list);
	XtSetSensitive(vdp->load.buttons[PLAYER_PLAY].widget, play);
	XtSetSensitive(vdp->load.buttons[PLAYER_STOP].widget, stop);
	XtSetSensitive(vdp->load.buttons[PLAYER_FF].widget, ff);
	XtSetSensitive(vdp->load.buttons[PLAYER_REW].widget, rew);
	XtSetSensitive(vdp->menus[LOAD_ATTACH].widget, open);
	XtSetSensitive(vdp->menus[LOAD_DETACH].widget, close);
}
